# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_arg_decl`
Contains the definition of the class that represents the lexing rule
to tokenize the declaration of an argument in a unit declaration.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import ARG_SYM, extract_identifier


class RuleArgDecl(LexingRule):
    def _apply_strategy(self, **kwargs):
        if not self._text.startswith(ARG_SYM, self._next_index):
            self.error_msg = \
                "Invalid token. Expected an argument declaration there " + \
                "(starting with '" + ARG_SYM + "')."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(
            LexicalToken(TerminalType.arg_marker, ARG_SYM)
        )

        arg_name = extract_identifier(self._text, self._next_index)
        if arg_name is None:
            self.error_msg = \
                "Didn't expect the line to end there. Expected an argument name."
            return False
        elif len(arg_name) == 0:
            self.error_msg = \
                "Couldn't extract the argument name. Arguments must have a name."
            return False
        self._next_index += len(arg_name)
        self._update_furthest_matched_index()
        self._tokens.append(LexicalToken(TerminalType.arg_name, arg_name))

        return True



#!/usr/bin/env python3
# coding: utf-8
"""
Module `chatette`
A generator of example sentences based on templates.
"""

# Retrieve the version number using setuptools
import pkg_resources
try:
    __version__ = pkg_resources.require("chatette")[0].version
except pkg_resources.DistributionNotFound:
    __version__ = "<couldn't retrieve version number>"  # TODO: find another way



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_variation`
Contains the definition of the class that represents the lexing rule
to tokenize a variation (in a unit declaration or reference).
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import VARIATION_SYM, extract_identifier


class RuleVariation(LexingRule):
    def _apply_strategy(self, **kwargs):
        if not self._text.startswith(VARIATION_SYM, self._next_index):
            self.error_msg = \
                "Invalid token. Expected a variation there (starting with '" + \
                VARIATION_SYM + "')."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(
            LexicalToken(TerminalType.variation_marker, VARIATION_SYM)
        )

        variation_name = extract_identifier(self._text, self._next_index)
        if variation_name is None:
            self.error_msg = \
                "Didn't expect an end of line there. Expected a variation name."
            return False
        elif len(variation_name) == 0:
            self.error_msg = \
                "Couldn't extract the name of the variation. Variation names " + \
                "must be at least one character long."
            return False

        self._next_index += len(variation_name)
        self._update_furthest_matched_index()
        self._tokens.append(
            LexicalToken(TerminalType.variation_name, variation_name)
        )

        return True



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_unit_decl`
Contains the definition of the class that represents the lexing rule
to tokenize a declaration of a unit (from '[' to ']').
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import \
    CASE_GEN_SYM, UNIT_END_SYM, extract_identifier

from chatette.parsing.lexing.rule_unit_start import RuleUnitStart
from chatette.parsing.lexing.rule_arg_decl import RuleArgDecl
from chatette.parsing.lexing.rule_variation import RuleVariation


class RuleUnitDecl(LexingRule):
    def _apply_strategy(self, **kwargs):
        if not self._try_to_match_rule(RuleUnitStart):
            return False

        if self._text.startswith(CASE_GEN_SYM, self._next_index):
            self._next_index += 1
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(TerminalType.casegen_marker, CASE_GEN_SYM)
            )

        identifier = extract_identifier(self._text, self._next_index)
        if identifier is not None:
            self._next_index += len(identifier)
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(TerminalType.unit_identifier, identifier)
            )

        if not self._match_any_order([None, RuleArgDecl, RuleVariation]):
            return False

        if not self._text.startswith(UNIT_END_SYM, self._next_index):
            self.error_msg = \
                "Invalid token. Expected the end of the unit declaration " + \
                "there (using symbol '" + UNIT_END_SYM + "')."
            return False

        # TODO maybe making a function for this would be useful
        if self._tokens[0].type == TerminalType.alias_decl_start:
            unit_end_type = TerminalType.alias_decl_end
        elif self._tokens[0].type == TerminalType.slot_decl_start:
            unit_end_type = TerminalType.slot_decl_end
        elif self._tokens[0].type == TerminalType.intent_decl_start:
            unit_end_type = TerminalType.intent_decl_end
        else:  # Should never happen
            raise ValueError(
                "An unexpected error happened during parsing: tried to " + \
                "parse the end of a unit but couldn't find its start in " + \
                "the previously parsed data.\nData was: " + str(self._tokens)
            )

        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(LexicalToken(unit_end_type, UNIT_END_SYM))

        return True



# coding: utf-8
"""
Module `chatette.units.modifiable.definitions.intent`
Contains the class representing an intent definition.
"""

from random import shuffle

from chatette.utils import UnitType
from chatette.units import IntentExample, add_example_no_dup
from chatette.units.modifiable.definitions.unit_definition import \
    UnitDefinition

from chatette.parsing import utils as putils


class IntentDefinition(UnitDefinition):
    """Represents an intent definition."""
    unit_type = UnitType.intent
    def __init__(
        self, identifier, modifiers,
        nb_training_examples=None, nb_testing_examples=None
    ):
        super(IntentDefinition, self).__init__(identifier, modifiers)
        self._nb_training_ex_asked = nb_training_examples
        self._nb_testing_ex_asked = nb_testing_examples

    def _compute_full_name(self):
        return "intent '" + self._name + "'"

    def set_nb_examples_asked(self, nb_training_ex, nb_testing_ex):
        """
        Sets the number of examples for this intent
        that should be in the training and test set.
        """
        self._nb_training_ex_asked = nb_training_ex
        self._nb_testing_ex_asked = nb_testing_ex

    def get_nb_training_examples_asked(self):
        return self._nb_training_ex_asked
    def get_nb_testing_examples_asked(self):
        return self._nb_testing_ex_asked


    def _make_empty_example(self):
        return IntentExample(self._name)

    def _example_to_intent_example(self, example):
        """
        Turns the example `example` (of type `Example`) into an instance of
        `IntentExample`.
        """
        return IntentExample.from_example(example, self._name)


    def _generate_random_strategy(self, variation_name=None):
        example = \
            super(IntentDefinition, self) \
                ._generate_random_strategy(variation_name=variation_name)
        return self._example_to_intent_example(example)

    def _generate_all_strategy(self, variation_name=None):
        examples = \
            super(IntentDefinition, self) \
                ._generate_all_strategy(variation_name=variation_name)
        return [self._example_to_intent_example(ex) for ex in examples]


    def generate_train(self):
        """
        Returns a list of examples to make up the training set.
        The list has as many examples as were asked in teh templates.
        """
        if self._nb_training_ex_asked is None:
            return self.generate_all()
        if self._nb_training_ex_asked == 0:
            return []
        return self.generate_nb_possibilities(self._nb_training_ex_asked)

    def generate_test(self, training_examples):
        """
        Returns a list of examples that can be put in the test set
        (not present in the training set).
        The list has as many examples as were asked in the templates.
        """
        if self._nb_testing_ex_asked is None or self._nb_testing_ex_asked == 0:
            return []
        if (
            self._nb_testing_ex_asked < \
            float(self.get_max_nb_possibilities()) / 5.0
        ):
            test_examples = []
            loop_count = 0
            while len(test_examples) < self._nb_testing_ex_asked:
                loop_count += 1
                current_ex = self.generate_random()
                if current_ex in training_examples:
                    continue
                add_example_no_dup(test_examples, current_ex)

                if loop_count > 10*self._nb_testing_ex_asked:
                    break
            return test_examples
        else:
            test_examples = []
            all_examples = self.generate_all()
            shuffle(all_examples)
            for ex in all_examples:
                if ex in training_examples:
                    continue
                add_example_no_dup(test_examples, ex)

                if len(test_examples) == self._nb_testing_ex_asked:
                    break
            return test_examples


    def as_template_str(self):
        result = ""
        for variation_name in self._variation_rules:
            result += putils.INTENT_SYM
            result += putils.UNIT_START_SYM
            result += putils.get_template_pre_modifiers(self._modifiers_repr)
            result += self._name
            result += putils.get_template_post_modifiers(self._modifiers_repr)
            if variation_name is not None:
                result += putils.VARIATION_SYM + variation_name
            result += putils.UNIT_END_SYM
            result += putils.ANNOTATION_START
            if self._nb_training_ex_asked is not None:
                result += \
                    putils.KEY_VAL_ENCLOSERS[1] + "training" + \
                    putils.KEY_VAL_ENCLOSERS[1] + \
                    putils.KEY_VAL_CONNECTOR + ' '
                result += str(self._nb_training_ex_asked)
                if self._nb_testing_ex_asked is not None:
                    result += putils.ANNOTATION_SEP
                    result += \
                        putils.KEY_VAL_ENCLOSERS[1] + "testing" + \
                        putils.KEY_VAL_ENCLOSERS[1] + \
                        putils.KEY_VAL_CONNECTOR + ' '
                    result += str(self._nb_testing_ex_asked)
            result += putils.ANNOTATION_END

            for rule in self._variation_rules[variation_name]:
                result += '\n\t' + rule.as_template_str()
        return result



#!/usr/bin/env python3
# coding: utf-8
"""
Module `chatette.utils`
Contains utility functions and classes used everywhere in the project.
"""

from __future__ import print_function
import sys
from random import sample, choice
from copy import deepcopy

from string import ascii_letters

from enum import Enum


class UnitType(Enum):
    alias = "alias"
    slot = "slot"
    intent = "intent"


class Singleton(object):
    """
    The base class for all singleton objects.
    Every class that subclasses this class will have the behavior
    of a singleton: their constructor will always return the same instance.
    @pre: In order to work, a sub-class needs to have an `_instance` class
          variable.
    """
    _instance = None
    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(Singleton, cls).__new__(cls)
        return cls._instance

    @classmethod
    def get_or_create(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = cls(*args, **kwargs)
        return cls._instance

    @classmethod
    def reset_instance(cls, *args, **kwargs):
        """
        Completely resets the instance of the class
        (representing the singleton), makes a new one and
        returns this instance.
        """
        cls._instance = None
        cls._instance = cls(*args, **kwargs)
        return cls._instance

    @classmethod
    def was_instantiated(cls):
        return (cls._instance is not None)


def cast_to_unicode(anything):
    """
    If executed with Python 2.7, cast any string that is in `anything`
    to unicode.
    If executed with Python 3, returns `anything`.
    `anything` can be a string, an array, a dict,...
    """
    if sys.version_info[0] == 3:
        return anything

    if isinstance(anything, str):
        return unicode(anything, "utf-8")
    if isinstance(anything, dict):
        cast_dict = dict()
        for key in anything:
            cast_key = cast_to_unicode(key)
            cast_value = cast_to_unicode(anything[key])
            cast_dict[cast_key] = cast_value
        return cast_dict
    if isinstance(anything, list):
        cast_list = []
        for e in anything:
            cast_list.append(cast_to_unicode(e))
        return cast_list
    return anything


def sample_indulgent(array, nb_items):
    """
    Same as `random.sample` but doesn't raise an error if `nb_items`
    is larger than the length of `array`: in that case,
    simply returns (a copy of) the whole array.
    """
    if nb_items <= len(array):
        return sample(array, nb_items)
    return deepcopy(array)


def rchop(string, ending):
    """Removes a substring at the end of a string."""
    if string.endswith(ending):
        return string[:-len(ending)]
    return string


def str_to_bool(text):
    """
    Transforms the strings 'True' and 'False' to their boolean counterparts.
    Raises a `ValueError` if `text` is neither of them.
    """
    text = text.lower()
    if text == "true":
        return True
    if text == "false":
        return False
    raise ValueError("Cannot convert '" + str(text) + "' into a boolean")


def remove_duplicates(dict_of_lists):
    """Removes duplicates from a dictionary containing lists."""
    return {key: list(set(value)) for (key, value) in dict_of_lists.items()}

def min_if_exist(n1, n2):
    """
    Returns the minimum between two numbers, or the only defined number
    (in case the other is `None`) or `None` if none of the numbers are defined.
    """
    if n1 is None and n2 is None:
        return None
    elif n1 is None:
        return n2
    elif n2 is None:
        return n1
    return min(n1, n2)


def random_string(length=6):
    """
    Returns a random string of length `length` containing only ASCII letters.
    """
    return ''.join([choice(ascii_letters) for _ in range(length)])


def append_to_list_in_dict(dict_of_lists, key, value):
    """
    Given the dict of key->lists `dict_of_lists`, appends the value `value` to
    the list at key `key` if this list exist. Otherwise, creates a list
    containing `value` and puts it at key `key`.
    """
    if key not in dict_of_lists:
        dict_of_lists[key] = [value]
    else:
        dict_of_lists[key].append(value)

def extend_list_in_dict(dict_of_lists, key, values):
    """
    Given the dict of key->lists `dict_of_lists`, extends the list at key `key`
    with the values `values` if this list exist. Otherwise, puts the list
    `values` at key `key`.
    """
    if key not in dict_of_lists:
        dict_of_lists[key] = values
    else:
        dict_of_lists[key].extend(values)


if __name__ == "__main__":
    # pylint: disable=wrong-import-position
    import warnings

    warnings.warn(
        "You are running the wrong file ('utils.py')." +
        "The file that should be run is 'run.py'."
    )



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_line`
Contains the class representing a lexing rule that applies to a full line.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing.rule_comment import RuleComment
from chatette.parsing.lexing.rule_file_inclusion import \
    RuleFileInclusion
from chatette.parsing.lexing.rule_unit_decl_line import \
    RuleUnitDeclLine
from chatette.parsing.lexing.rule_unit_rule import RuleUnitRule


class RuleLine(LexingRule):
    """Represents the lexing rule for a full line of a template file."""
    _empty_match_allowed = True
    def __init__(self, text):
        super(RuleLine, self).__init__(text, 0)

    def _apply_strategy(self, **kwargs):
        if self._match_one_of(
            [RuleComment, RuleFileInclusion, RuleUnitDeclLine, RuleUnitRule],
            self._next_index,
            **kwargs
        ):
            if self._next_index < len(self._text):
                comment_rule = RuleComment(self._text, self._next_index)
                if not comment_rule.matches():
                    self.error_msg = "Invalid token. Expected a comment or " + \
                        "the end of the line there."
                    self._update_furthest_matched_index(comment_rule)
                    return False
                self._tokens.extend(comment_rule.get_lexical_tokens())
                self._next_index = comment_rule.get_next_index_to_match()
                self._update_furthest_matched_index()
                # Comments end the line BY DESIGN
                return True

            return True

        return False



"""
Module `chatette.parsing.line_count_file_wrapper`.
Contains a wrapper of `io.File` that counts on which line it is currently.
"""

import io

from chatette.utils import cast_to_unicode


class LineCountFileWrapper(object):
    """
    A wrapper of `io.File` that keeps track of the line number it is reading.
    """

    def __init__(self, filepath, mode='r'):
        self.name = cast_to_unicode(filepath)
        self.f = io.open(filepath, mode)
        self.line_nb = 0

    def close(self):
        return self.f.close()
    def closed(self):
        return self.f.closed

    def readline(self):
        self.line_nb += 1
        return self.f.readline()

    # to allow using in 'with' statements
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.f.close()
        self.close()




from sys import version_info


def _get_python_version_as_str():
    return str(version_info[0]) + '.' + str(version_info[1])


def _is_supported_python_version():
    return version_info[0] == 3 \
        or version_info[0] == 2 and version_info[1] == 7

def _is_deprecated_python_version():
    return version_info[0] == 2 \
        or version_info[0] == 3 and version_info[1] < 4



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_word`
Contains the definition of the class that represents the lexing rule
to tokenize a word (inside a rule).
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import \
    FILE_INCLUSION_SYM, \
    UNIT_START_SYM, UNIT_END_SYM, \
    ALIAS_SYM, SLOT_SYM, INTENT_SYM, \
    SLOT_VAL_SYM, \
    CHOICE_START, CHOICE_END, CHOICE_SEP, \
    OLD_CHOICE_START, OLD_CHOICE_END, OLD_CHOICE_SEP, \
    RAND_GEN_SYM, \
    find_unescaped, find_next_comment
from chatette.utils import min_if_exist


class RuleWord(LexingRule):
    _should_be_escaped_chars = [
        FILE_INCLUSION_SYM,
        UNIT_START_SYM, UNIT_END_SYM,
        ALIAS_SYM, SLOT_SYM, INTENT_SYM,
        CHOICE_START, CHOICE_END,
        OLD_CHOICE_START, OLD_CHOICE_END
    ]
    _should_be_escaped_in_choices_chars = [
        CHOICE_SEP,
        OLD_CHOICE_SEP,
        RAND_GEN_SYM
    ]
    _should_be_escaped_in_slot_def_chars = [SLOT_VAL_SYM]

    def _apply_strategy(self, **kwargs):
        """
        `kwargs` can contain a boolean with key `inside_choice` that is
        `True` when the current word is inside a choice and `False` otherwise.
        If this boolean is not in `kwargs`, defaults to `False`.
        ´kwargs´ can also contain a boolean with key `parsing_slot_def`
        which is `True` iff the current is in a rule inside a slot definition.
        If this boolean is not in `kwargs`, defaults to `False`.
        """
        inside_choice = kwargs.get("inside_choice", False)
        parsing_slot_def = kwargs.get("parsing_slot_def", False)

        # TODO this might be better using regexes
        if self._text[self._start_index].isspace():
            self.error_msg = \
                "Invalid token. Expected a word instead of a whitespace there."
            return False

        # Find whitespace after the word
        next_word_index = self._start_index + 1  # NOTE exclusive
        while True:
            if (
                next_word_index == len(self._text)
                or self._text[next_word_index].isspace()
            ):
                break
            next_word_index += 1

        next_word_index = \
            min_if_exist(
                next_word_index,
                find_next_comment(self._text, self._start_index)
            )

        if next_word_index == self._start_index:
            self.error_msg = "Invalid token. Expected a word to start here."
            return False
        for current_char in RuleWord._should_be_escaped_chars:
            if next_word_index == self._start_index:
                break
            next_word_index = \
                min_if_exist(
                    next_word_index,
                    find_unescaped(self._text, current_char, self._start_index)
                )

        if inside_choice and next_word_index > self._start_index:
            for char_to_escape in RuleWord._should_be_escaped_in_choices_chars:
                next_word_index = \
                    min_if_exist(
                        next_word_index,
                        find_unescaped(
                            self._text, char_to_escape, self._start_index
                        )
                    )

        if parsing_slot_def and next_word_index > self._start_index:
            for char_to_escape in RuleWord._should_be_escaped_in_slot_def_chars:
                next_word_index = \
                    min_if_exist(
                        next_word_index,
                        find_unescaped(
                            self._text, char_to_escape, self._start_index
                        )
                    )

        if next_word_index == self._start_index:
            self.error_msg = "Invalid token. Expected a word to start here."
            return False

        word = self._text[self._start_index:next_word_index]
        self._next_index = next_word_index
        self._update_furthest_matched_index()
        self._tokens.append(LexicalToken(TerminalType.word, word))
        return True



# coding: utf-8
"""
Module `chatette.parsing.lexing.lexing_rule`
Contains the class that represents a lexing rule, which is
in charge of checking if the rule it represents matces the input,
transforms it to lexical tokens (simply called tokens in the code) and
formulates the error if required. Therefore, each instance of such a class is
stateful.
"""

from abc import ABCMeta, abstractmethod
from future.utils import with_metaclass

from chatette.parsing.input_file_manager import InputFileManager


class LexingRule(with_metaclass(ABCMeta, object)):
    """
    Superclass for all lexing rules.
    Child classes each represent a concrete lexing rule (each instance keeping
    a state) and whose purpose will be to:
    - check whether the provided text matches the rule it represents
    - produce the sequence of lexical toekns if it matches
    - formulate the error and communicate it to the user if asked
    """
    _empty_match_allowed = False
    def __init__(self, text, start_index):
        self._text = text

        self._start_index = start_index
        self._next_index = start_index
        self._furthest_matched_index = start_index  # inclusive
        self._tokens = []

        self._matched = None
        self.error_msg = None


    def matches(self, **kwargs):
        """
        Applies the rule to `self._text`
        and returns `True` if the rule could be applied successfully.
        `kwargs` will be passed to the subsequent call to `self._apply_strategy`.
        @post: `self.get_lexical_tokens()` will return valid tokens
               iff this method returned `True`.
        """
        if self._matched is None:
            if self._start_index >= len(self._text):
                if self.__class__._empty_match_allowed:
                    self._matched = True
                else:
                    self._matched = False
                    self.error_msg = \
                        "Matching of rule '" + self.__class__.__name__ + "' " + \
                        "failed. Didn't expect an end of line there."
            else:
                self._matched = self._apply_strategy(**kwargs)
        return self._matched
    @abstractmethod
    def _apply_strategy(self, **kwargs):
        """
        Strategy to apply the rule to the text `self._text`.
        Implements the design pattern "Strategy method".
        Returns `True` iff the rule could be applied successfully.
        If the rule doesn't match, `self.error_msg` should be updated
        accordingly.
        `kwargs` can be used to pass arguments that are specific to
        the current strategy.
        After this method was called, the following variables should be
        up-to-date:
        - if it matched: `self._tokens`, `self._next_index`,
                         `self._furthest_matched_index`
        - if it didn't match: `self._furthest_matched_index`, `self.error_msg`
        """
        raise NotImplementedError()

    # Helpers for matching rules
    def _try_to_match_rule(self, rule_class, index=None, **kwargs):
        """
        Tries to match the rule represented by `rule_class`
        and updates `self._tokens` and `self._next_index` if the rule could be
        applied. If the rule couldn't be applied, `self.error_msg` is updated.
        `rule_class` is a subclass of `LexingRule`; `index` is an index in the
        text `self._text`.
        `kwargs` will be passed down to the subsequent `_matches`
        @returns: `True` iff the rule could be applied successfully.
        """
        if index is None:
            index = self._next_index
        rule = rule_class(self._text, index)
        if rule.matches(**kwargs):
            self._tokens.extend(rule.get_lexical_tokens())
            self._next_index = rule.get_next_index_to_match()
            self._update_furthest_matched_index(rule)
            return True
        else:
            self.error_msg = rule.error_msg
            self._update_furthest_matched_index(rule)
            return False

    def _match_one_of(self, rule_classes, index=None, **kwargs):
        """
        Finds which of the rules in `rule_classes` matches the text starting at
        index `index`.
        Try rules in order of appearance in `rule_classes`.
        `rule_classes` is a list of class of rules.
        Returns `True` iff one of the rules could apply.
        `kwargs` will be passed down to the subsequent `_matches`
        @pre: `rule_classes` contains at least one element.
        @post:
            - if a rule matched, `self._tokens` and `self._next_index`
              are updated.
            - if no rule matched, `self._matched` and `self.error_msg` are
              updated.
        @raises: `ValueError` if `rule_classes` does not contain any element.
        """
        if len(rule_classes) == 0:
            raise ValueError(
                "Lexer tried to apply no rule at all " + \
                "during application of rule '" + self.__class__.__name__ + \
                "'."
            )
        if index is None:
            index = self._next_index

        best_failed_rule = None
        longest_match_last_index = None
        for rule_class in rule_classes:
            rule = rule_class(self._text, index)
            if rule.matches(**kwargs):
                self._tokens.extend(rule.get_lexical_tokens())
                self._next_index = rule.get_next_index_to_match()
                return True
            else:
                match_last_index = rule._furthest_matched_index
                # match_last_index = rule.get_next_index_to_match() - index
                if best_failed_rule is None or match_last_index > longest_match_last_index:
                    best_failed_rule = rule
                    longest_match_last_index = match_last_index

        self._update_furthest_matched_index(best_failed_rule)

        # No rule matched
        self._matched = False
        if longest_match_last_index - index == 0:
            self.error_msg = \
                "Invalid token. Expected either of the following rules: '" + \
                "', '".join([c.__name__ for c in rule_classes]) + "'."
        else:
            self.error_msg = best_failed_rule.error_msg
        return False

    def _match_any_order(self, rule_classes, index=None):
        """
        Matches as many rules as possible from `rule_classes` and this in any
        order, but each rule can be matched only once.
        If `None` is present in `rule_classes`, it is allowed to match 0 rules.
        Returns `True` if at least one rule matched (or 0 if allowed).
        `rule_classes` is a list of class of rules.
        @pre: `rule_classes` contains at least one element.
        @post:
            - if at least one rule matched, `self._tokens` and
              `self._next_index` are updated.
            - if no rule matched and at least 1 must be matched,
              `self._matched` and `self.error_msg` are updated.
            - if no rule matched and 0 rules were allowed to be matched,
              simply returns `False`. `self._matched` and `self.error_msg` are
              NOT updated.
        @raises: `ValueError` if `rule_classes` does not contain any element.
        """
        if len(rule_classes) == 0:
            raise ValueError(
                "Lexer tried to apply no rule at all " + \
                "during application of rule '" + self.__class__.__name__ + \
                "'."
            )
        if index is None:
            index = self._next_index
        no_match_allowed = False
        matched_some_rule = False

        remaining_rules = list(rule_classes)
        last_nb_remaining_rules = len(remaining_rules)  # Checked everytime every rules have been tested
        i = 0
        best_failed_rule = None
        longest_match_last_index = None
        while True:
            if i >= len(remaining_rules):
                if (
                    len(remaining_rules) == 0
                    or len(remaining_rules) == last_nb_remaining_rules
                ):
                    break
                i = 0
                last_nb_remaining_rules = len(remaining_rules)
            if remaining_rules[i] is None:
                no_match_allowed = True
                remaining_rules.remove(None)
                i = 0
            else:
                rule = (remaining_rules[i])(self._text, index)
                if rule.matches():
                    matched_some_rule = True
                    self._tokens.extend(rule.get_lexical_tokens())
                    self._next_index = rule.get_next_index_to_match()
                    index = self._next_index
                    remaining_rules.remove(remaining_rules[i])
                else:
                    i += 1
                    match_last_index = rule._furthest_matched_index
                    # match_last_index = rule.get_next_index_to_match() - index
                    if best_failed_rule is None or match_last_index > longest_match_last_index:
                        best_failed_rule = rule
                        longest_match_last_index = match_last_index

        self._update_furthest_matched_index(best_failed_rule)

        if matched_some_rule:
            return True
        if no_match_allowed:
            return True
        # No rule matched and no match is not allowed
        self._matched = False
        if longest_match_last_index - index == 0:
            self.error_msg = \
                "Invalid token. Expected either of the following rules: '" + \
                "', '".join([c.__name__ for c in rule_classes]) + "'."
        else:
            self.error_msg = best_failed_rule.error_msg
        return False


    # Setters
    def _update_furthest_matched_index(self, rule=None):
        """
        Updates the value of `self._furthest_matched_index`
        so that it is the maximum between its current value and
        the value of `rule._furthest_matched_index`.
        If `rule` is `None`, updates `self._furthest_matched_index`
        with respect to `self._next_index` (i.e. the index that should
        be tested for a match next).
        The value of `self._furthest_matched_index` represents
        the maximum index that was matched by a rule
        called by the current rule (or by this rule itself).
        """
        if rule is None:
            self._furthest_matched_index = \
                max(self._furthest_matched_index, self._next_index - 1)
        else:
            self._furthest_matched_index = \
                max(self._furthest_matched_index, rule._furthest_matched_index)

    # Getters
    def get_lexical_tokens(self):
        """
        Returns the lexical tokens after th rule was successfully applied
        using `self.matches()`.
        If this method is called before the rule was applied,
        the rule will be applied, but a `ValueError` may be raised in case
        the rule doesn't match the text.
        @returns: a list of `LexicalToken`s
        """
        if self._matched is None:
            self.matches()
        if not self._matched:
            raise ValueError(
                "Asked for the lexical tokens of a rule that did not " + \
                "match: '" + self.__class__.__name__ + "'."
            )
        return self._tokens
    def get_next_index_to_match(self):
        """
        Returns the next index in `self._text` that corresponds to
        the rule next in line. This index may correspond to the end of the text
        if the current rule covered the whole text or to the same index as
        `self._start_index` if the rule didn't match anything at all.
        @returns: an integer index
        """
        if self._matched is None:
            self.matches()
        return self._next_index


    def print_error(self):
        """
        If the corresponding rule didn't match,
        prints the relevant error so the user is aware of what the lexing
        problem was.
        """
        if self._matched is None:
            raise ValueError(
                "Tried to print an error for rule '" + self.__class__.__name__ + \
                "' before it was applied."
            )
        if self._matched:
            raise ValueError(
                "Tried to print an error for rule '" + self.__class__.__name__ + \
                "' when no error occurred."
            )
        if self.error_msg is None:
            raise ValueError(
                "No error to print for rule '" + self.__class__.__name__ + \
                "'."
            )
        InputFileManager.get_or_create() \
            .syntax_error(self.error_msg, self._furthest_matched_index + 1)



"""
Module `chatette.cli.interactive_commands.rename_command`.
Contains the strategy class that represents the interactive mode command
`rename` which changes the name of a unit (if it exists).
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.units.ast import AST


class RenameCommand(CommandStrategy):

    def execute(self):
        """
        Implements the command `rename` which renames a unit
        into something else. Displays an error if the unit wasn't found.
        """
        if len(self.command_tokens) < 4:
            self.print_wrapper.error_log(
                "Missing some arguments\nUsage: " + \
                'rename <unit-type> "<old-name>" ' + '"<new-name>"')
            return
        unit_type = \
            CommandStrategy.get_unit_type_from_str(self.command_tokens[1])

        if unit_type is None:
            self.print_wrapper.error_log(
                "Unknown unit type: '" + str(self.command_tokens[1]) + "'."
            )
        else:
            old_name = CommandStrategy.remove_quotes(self.command_tokens[2])
            new_name = CommandStrategy.remove_quotes(self.command_tokens[3])
            if new_name == "":
                self.print_wrapper.error_log(
                    "An empty name is not a valid " + unit_type.name + " name."
                )
                return

            try:
                AST.get_or_create().rename_unit(unit_type, old_name, new_name)
                self.print_wrapper.write(
                    unit_type.name.capitalize() + " '" + old_name + \
                    "' was successfully renamed to '" + new_name + "'.")
            except KeyError:
                self.print_wrapper.error_log(
                    "Couldn't find a unit named '" + str(old_name) + "'."
                )
            except ValueError:
                self.print_wrapper.error_log(
                    unit_type.name.capitalize() + " '" + new_name + \
                    "' is already in use."
                )


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



# coding: utf-8

"""
Module `chatette.units.modifiable.definitions.alias`
Contains the class representing an alias definition.
"""

from chatette.utils import UnitType
from chatette.units.modifiable.definitions.unit_definition import \
    UnitDefinition


class AliasDefinition(UnitDefinition):
    """Represents an alias definition."""
    unit_type = UnitType.alias
    def _compute_full_name(self):
        return "alias '" + self._name + "'"



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_whitespaces`
Contains the class representing the lexing rule that applies to whitespaces
and indentation.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType


class RuleWhitespaces(LexingRule):
    def _apply_strategy(self, **kwargs):
        """
        `kwargs` can contain a boolean at key `parsing_indentation` that is
        `True` iff the whitespaces currently parsed correspond to
        an indentation, and `False` if it corresponds to simple whitespaces.
        If `kwargs` doesn't contain this boolean, defaults to `False`.
        """
        parsing_indentation = kwargs.get("parsing_indentation", False)
        if parsing_indentation:
            terminal_type = TerminalType.indentation
        else:
            terminal_type = TerminalType.whitespace
        text = self._text

        while self._next_index < len(text) and text[self._next_index].isspace():
            self._next_index += 1
            self._update_furthest_matched_index()
        if self._next_index > self._start_index:
            matched_text = text[self._start_index:self._next_index]
            self._tokens.append(LexicalToken(terminal_type, matched_text))
            return True

        self.error_msg = "Invalid token. Expected at least one whitespace there."
        return False



# coding: utf-8

from warnings import warn

from chatette.utils import Singleton, cast_to_unicode
from chatette.log import print_warn
from chatette.parsing.utils import \
    OLD_COMMENT_SYM, COMMENT_SYM, \
    OLD_CHOICE_START, OLD_CHOICE_END, CHOICE_START, CHOICE_END


class Deprecations(Singleton):
    _instance = None
    def __init__(self):
        self._old_comment_warned = False
        self._old_choice_warned = False

    def warn_old_comment(self, filename=None, line_nb=None, line=None):
        """
        Warns the user on stderr that one of their files contains semicolons
        comments (which are a deprecated way of making comments).
        Rather use '//' comments instead of ';' comments.
        """
        if not self._old_comment_warned:
            self._old_comment_warned = True
            message = \
                "Comments starting with a semi-colon '" + \
                OLD_COMMENT_SYM + "' are now deprecated. " + \
                "Please use the new double slash '" + COMMENT_SYM + \
                "' syntax instead."
            if filename is not None:
                message += \
                    "\nThis syntax was found in file '" + \
                    cast_to_unicode(filename) + "'"
                if line_nb is not None and line is not None:
                    message += \
                        " at line " + str(line_nb) + ": '" + \
                        str(line).strip() + "'"
                message += '.'
            elif line_nb is not None and line is not None:
                message += \
                    "\nThis syntax was found at line " + str(line_nb) + \
                    ": '" + str(line).strip() + "'."
            warn(message, DeprecationWarning)
            print_warn(message)

    def warn_old_choice(self, filename=None, line_nb=None, line=None):
        """
        Warns the user on stderr that one of their files contains semicolons
        comments (which are a deprecated way of making comments).
        Rather use '//' comments instead of ';' comments.
        """
        if not self._old_choice_warned:
            self._old_choice_warned = True
            message = \
                "Choices starting with '" + OLD_CHOICE_START + \
                "' and ending with '" + OLD_CHOICE_END + \
                "' are now deprecated. Please use the new syntax that " + \
                "starts with '" + CHOICE_START + "' and ends with '" + \
                CHOICE_END + "' instead."
            if filename is not None:
                message += \
                    "\nThis syntax was found in file '" + \
                    cast_to_unicode(filename) + "'"
                if line_nb is not None and line is not None:
                    message += \
                        " at line " + str(line_nb) + ": '" + \
                        str(line).strip() + "'"
                message += '.'
            elif line_nb is not None and line is not None:
                message += \
                    "\nThis syntax was found at line " + str(line_nb) + \
                    ": '" + str(line).strip() + "'."
            warn(message, DeprecationWarning)
            print_warn(message)



"""
Module `chatette.cli.interactive_commands.show_command`.
Contains the strategy class that represents the interactive mode command
`show` which shows information about a unit definition and lists a bunch of
its rules (all if possible).
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.units.ast import AST


class ShowCommand(CommandStrategy):
    usage_str = 'show <unit-type> "<unit-name>"'
    max_nb_rules_to_display = 12

    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        try:
            unit = AST.get_or_create()[unit_type][unit_name]
            self.print_wrapper.write(unit.short_description())

            if variation_name is None:
                self.print_wrapper.write(
                    "Template rules:\n" + str(unit.as_template_str())
                )
            else:  # TODO
                pass
        except KeyError:
            self.print_wrapper.write(
                unit_type.name.capitalize() + " '" + unit_name + \
                "' is not defined."
            )



# coding: utf-8
"""
Module `chatette.units.modifiable.unit_reference`
Contains a class representing all the references to unit definition
that are present in template rules.
"""

from collections import OrderedDict

from chatette.units.modifiable import ModifiableItem
from chatette.units.ast import AST

from chatette.parsing import utils as putils


class UnitReference(ModifiableItem):
    """
    Represents a reference to a unit definition that can be contained
    in a template rule.
    """
    def __init__(self, identifier, unit_type, leading_space, modifiers):
        self._unit_type = unit_type
        super(UnitReference, self).__init__(
            identifier, leading_space, modifiers
        )
        try:
            self._definition = AST.get_or_create()[unit_type][identifier]
            self._fix_arg_value_modifier()
        except KeyError:
            self._definition = None


    def _compute_full_name(self):
        return "reference to " + self._unit_type.value + " '" + \
            self._name + "'"


    def get_definition(self):
        if self._definition is None:
            try:
                self._definition = \
                    AST.get_or_create()[self._unit_type][self._name]
                self._fix_arg_value_modifier()
            except KeyError:
                raise KeyError(
                    "Couldn't find the definition corresponding to " + \
                    self.full_name + "."
                )
        return self._definition


    def _compute_nb_possibilities(self):
        return \
            self.get_definition() \
                .get_max_nb_possibilities(
                    variation_name=self._modifiers_repr.variation_name
                )

    def _generate_random_strategy(self):
        return \
            self.get_definition() \
                .generate_random(
                    variation_name=self._modifiers_repr.variation_name
                )

    def _generate_all_strategy(self):
        return \
            self.get_definition() \
                .generate_all(
                    variation_name=self._modifiers_repr.variation_name
                )

    def _generate_n_strategy(self, n):
        return \
            self.get_definition() \
                .generate_nb_possibilities(
                    n, variation_name=self._modifiers_repr.variation_name
                )


    def _fix_arg_value_modifier(self):
        """
        Transforms the modifier representation to have a mapping between
        an argument name and value rather than the value only.
        This method has no effect if:
            - no argument value is set.
            - a mapping is already set as this modifier.
            - the definition doesn't exist yet.
        @raises: - `SyntaxError` if the definition does not have an argument.  # TODO should it be just a warning?
                 - `SyntaxError` if the definition has more than one argument
                   defined.
        """
        if self._definition is not None and (
            self._modifiers_repr.argument_value is not None
            and not isinstance(self._modifiers_repr.argument_value, dict)
        ):
            arg_name = self._definition._modifiers_repr.argument_name
            if arg_name is None:
                raise SyntaxError(
                    self.full_name.capitalize() + " was given an argument " + \
                    "while the definition it references has no argument " + \
                    "defined."
                )
            if isinstance(arg_name, list):
                raise SyntaxError(
                    self.full_name.capitalize() + " was given a single " + \
                    "unnamed argument, while the definition " + \
                    "it references defines several arguments."
                )
            mapping = \
                OrderedDict(
                    [(arg_name, self._modifiers_repr.argument_value)]
                )
            self._modifiers_repr.argument_value = mapping

    def as_template_str(self):
        result = ""
        result += putils.get_template_unit_sym(self._unit_type)
        result += putils.UNIT_START_SYM
        result += putils.get_template_pre_modifiers(self._modifiers_repr)
        result += self._name
        result += putils.get_template_post_modifiers(self._modifiers_repr)
        result += putils.UNIT_END_SYM
        if self._leading_space:
            result = ' ' + result
        return result



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_slot_val`
Contains the definition of the class that represents the lexing rule
to tokenize a slot value being set within a unit rule (only for a slot).
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import find_next_comment, SLOT_VAL_SYM


class RuleSlotVal(LexingRule):
    def _apply_strategy(self, **kwargs):
        """
        `kwargs` can contain a boolean with key `parsing_slot_def` that is
        `True` if the current text is part of a slot definition.
        If this boolean is not in `kwargs`, defaults to `False`.
        """
        parsing_slot_def = kwargs.get("parsing_slot_def", False)
        if parsing_slot_def:
            while self._text[self._next_index].isspace():
                self._next_index += 1
                self._update_furthest_matched_index()

            if self._text.startswith(SLOT_VAL_SYM, self._next_index):
                self._tokens.append(
                    LexicalToken(TerminalType.slot_val_marker, SLOT_VAL_SYM)
                )
                self._next_index += 1
                self._update_furthest_matched_index()

                while self._text[self._next_index].isspace():
                    self._next_index += 1
                    self._update_furthest_matched_index()

                comment_sym = find_next_comment(self._text, self._next_index)
                if comment_sym is not None:
                    slot_value = \
                        self._text[self._next_index:comment_sym].rstrip()
                else:
                    slot_value = self._text[self._next_index:].rstrip()

                self._tokens.append(
                    LexicalToken(TerminalType.slot_val, slot_value)
                )
                self._next_index += len(slot_value)
                self._update_furthest_matched_index()

                return True

            return False
        else:
            raise ValueError(
                "Tried to extract a slot value within a rule that is not " + \
                "part of a slot definition."
            )



"""
Module `chatette.cli.interactive_commands.exit_command`.
Contains the strategy class that represents the interactive mode command
`exit` which exits the interactive mode.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy


class ExitCommand(CommandStrategy):

    def execute(self):
        pass

    def should_exit(self):
        return True


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



from .pyversion import _is_deprecated_python_version, _get_python_version_as_str
from chatette.log import print_warn


def check_for_deprecations():
    _check_python_version()

def _check_python_version():
    if _is_deprecated_python_version():
        print_warn(
            "Python v" + _get_python_version_as_str() + \
            " will not be supported in the future. " + \
            "Please upgrade Python whenever possible."
        )



#!/usr/bin/env python3
# coding: utf-8

"""
Module `chatette.units.word`
Contains the definition of words as contents of rules.
"""

from chatette.units import Example
from chatette.units.generating_item import GeneratingItem


class Word(GeneratingItem):
    """Represents a word as a content of a rule."""
    def __init__(self, name, leading_space):
        super(Word, self).__init__(name, leading_space)
        self.word = name

    def _compute_full_name(self):
        return "word '" + self._name + "'"

    def get_max_nb_possibilities(self):
        return 1
    def _compute_nb_possibilities(self):
        return 1

    def generate_random(self, **kwargs):
        return self._generate_random_strategy()
    def _generate_random_strategy(self):
        if self._leading_space:
            return Example(' ' + self._name)
        return Example(self._name)

    def generate_all(self):
        return self._generate_all_strategy()
    def _generate_all_strategy(self):
        if self._leading_space:
            return [Example(' ' + self._name)]
        return [Example(self._name)]

    def generate_nb_possibilities(self, nb_possibilities):
        return self._generate_n_strategy()
    def _generate_n_strategy(self, n=None):
        if self._leading_space:
            return [Example(' ' + self._name)]
        return [Example(self._name)]

    def as_template_str(self):
        if self._leading_space:
            return ' ' + self.word
        return self.word



# coding: utf-8
"""
Module `chatette.statistics`
Contains the singleton that is meant to record statistics about the execution
of the program.
"""

from chatette.utils import Singleton, UnitType


class InvalidStatsState(Exception):
    pass


class Stats(Singleton):
    _instance = None
    def __init__(self):
        # File counts
        self.nb_input_files_parsed = 0
        # self.nb_output_files = 0

        # Unit counts
        self.nb_units_declared = 0
        self.nb_intents_declared = 0
        self.nb_slots_declared = 0
        self.nb_aliases_declared = 0
        # Variation counts
        self.nb_variation_units = 0
        self.nb_variation_intents = 0
        self.nb_variation_slots = 0
        self.nb_variation_aliases = 0

        # Rule counts
        self.nb_rules_parsed = 0

    def __str__(self):
        return \
            "Statistics:\n\tParsed files: " + str(self.nb_input_files_parsed) + \
            "\n\tDeclared units: " + str(self.nb_units_declared) + \
            " (" + str(self.nb_variation_units) + " variations)" + \
            "\n\t\tDeclared intents: " + str(self.nb_intents_declared) + \
            " (" + str(self.nb_variation_intents) + " variations)" + \
            "\n\t\tDeclared slots: " + str(self.nb_slots_declared) + \
            " (" + str(self.nb_variation_slots) + " variations)" + \
            "\n\t\tDeclared aliases: " + str(self.nb_aliases_declared) + \
            " (" + str(self.nb_variation_aliases) + " variations)" + \
            "\n\tParsed rules: " + str(self.nb_rules_parsed)

    def new_file_parsed(self):
        self.nb_input_files_parsed += 1
    # def new_file_written(self):
    #     self.nb_output_files += 1

    def new_unit_declared(self, unit_type):
        if unit_type == UnitType.alias:
            self.new_alias_declared()
        elif unit_type == UnitType.slot:
            self.new_slot_declared()
        elif unit_type == UnitType.intent:
            self.new_intent_declared()
        else:
            raise TypeError(
                "Tried to increase the statistics for unit declarations " + \
                "with an unknown unit type (" + str(unit_type) + ")."
            )
    def new_intent_declared(self):
        self.nb_units_declared += 1
        self.nb_intents_declared += 1
    def new_slot_declared(self):
        self.nb_units_declared += 1
        self.nb_slots_declared += 1
    def new_alias_declared(self):
        self.nb_units_declared += 1
        self.nb_aliases_declared += 1

    def new_variation_unit_declared(self, unit_type):
        if unit_type == UnitType.alias:
            self.new_variation_alias_declared()
        elif unit_type == UnitType.slot:
            self.new_variation_slot_declared()
        elif unit_type == UnitType.intent:
            self.new_variation_intent_declared()
        else:
            raise TypeError(
                "Tried to increase the statistics for a variation of " + \
                "unit declarations with an unknown unit type (" + \
                str(unit_type) + ")."
            )
    def new_variation_intent_declared(self):
        self.nb_variation_units += 1
        self.nb_variation_intents += 1
    def new_variation_slot_declared(self):
        self.nb_variation_units += 1
        self.nb_variation_slots += 1
    def new_variation_alias_declared(self):
        self.nb_variation_units += 1
        self.nb_variation_aliases += 1

    def new_rule_parsed(self):
        self.nb_rules_parsed += 1
    def new_rules_parsed(self, nb_rules):
        self.nb_rules_parsed += nb_rules

    def one_unit_removed(self, unit_type):
        if unit_type == UnitType.alias:
            self.one_alias_removed()
        elif unit_type == UnitType.slot:
            self.one_slot_removed()
        elif unit_type == UnitType.intent:
            self.one_intent_removed()
        else:
            raise TypeError(
                "Tried to increase the statistics for a variation of " + \
                "unit declarations with an unknown unit type (" + \
                str(unit_type) + ")."
            )
    def one_intent_removed(self):
        self.nb_units_declared -= 1
        if self.nb_units_declared < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for unit declarations below 0."
            )
        self.nb_intents_declared -= 1
        if self.nb_intents_declared < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for intent declarations below 0."
            )
    def one_slot_removed(self):
        self.nb_units_declared -= 1
        if self.nb_units_declared < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for unit declarations below 0."
            )
        self.nb_slots_declared -= 1
        if self.nb_slots_declared < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for slot declarations below 0."
            )
    def one_alias_removed(self):
        self.nb_units_declared -= 1
        if self.nb_units_declared < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for unit declarations below 0."
            )
        self.nb_aliases_declared -= 1
        if self.nb_aliases_declared < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for alias declarations below 0."
            )

    def several_variation_units_removed(self, unit_type, nb_variations):
        for _ in range(nb_variations):
            self.one_variation_unit_removed(unit_type)
    def one_variation_unit_removed(self, unit_type):
        if unit_type == UnitType.alias:
            self.one_variation_alias_removed()
        elif unit_type == UnitType.slot:
            self.one_variation_slot_removed()
        elif unit_type == UnitType.intent:
            self.one_variation_intent_removed()
        else:
            raise TypeError(
                "Tried to increase the statistics for a variation of " + \
                "unit declarations with an unknown unit type (" + \
                str(unit_type) + ")."
            )
    def one_variation_intent_removed(self):
        self.nb_variation_units -= 1
        if self.nb_variation_units < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for variations of " + \
                "unit declarations below 0."
            )
        self.nb_variation_intents -= 1
        if self.nb_variation_intents < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for variations of " + \
                "intent declarations below 0."
            )
    def one_variation_slot_removed(self):
        self.nb_variation_units -= 1
        if self.nb_variation_units < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for variations of " + \
                "unit declarations below 0."
            )
        self.nb_variation_slots -= 1
        if self.nb_variation_slots < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for variations of " + \
                "slot declarations below 0."
            )
    def one_variation_alias_removed(self):
        self.nb_variation_units -= 1
        if self.nb_variation_units < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for variations of " + \
                "unit declarations below 0."
            )
        self.nb_variation_aliases -= 1
        if self.nb_variation_aliases < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for variations of " + \
                "alias declarations below 0."
            )

    def one_rule_removed(self):
        self.nb_rules_parsed -= 1
        if self.nb_rules_parsed < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for parsed rules below 0."
            )
    def several_rules_removed(self, nb_rules):
        self.nb_rules_parsed -= nb_rules
        if self.nb_rules_parsed < 0:
            raise InvalidStatsState(
                "Tried to decrement statistics for parsed rules below 0."
            )



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_rand_gen`
Contains the definition of the class that represents the lexing rule
to tokenize a randgen modifier (inside a unit declaration, reference or
choice).
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import \
    extract_identifier, \
    RAND_GEN_SYM, RAND_GEN_PERCENT_SYM, RAND_GEN_OPPOSITE_SYM

from chatette.parsing.lexing.rule_percent_gen import RulePercentGen


class RuleRandGen(LexingRule):
    def _apply_strategy(self, **kwargs):
        if not self._text.startswith(RAND_GEN_SYM, self._next_index):
            self.error_msg = \
                "Invalid token. Expected a random generation modifier to " + \
                "begin there (starting with '" + RAND_GEN_SYM + "')."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(
            LexicalToken(TerminalType.randgen_marker, RAND_GEN_SYM)
        )

        if self._text.startswith(RAND_GEN_OPPOSITE_SYM, self._next_index):
            self._next_index += 1
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(
                    TerminalType.opposite_randgen_marker, RAND_GEN_OPPOSITE_SYM
                )
            )

        # TODO not sure `extract_identifier` is the best thing to use here
        randgen_name = extract_identifier(self._text, self._next_index)
        if randgen_name is None:
            self.error_msg = \
                "Didn't expect the line to end there. Expected a name for " + \
                "the random generation modifier, a percentage for it or " + \
                "the end of the unit or choice."
            return False
        if len(randgen_name) > 0:
            self._next_index += len(randgen_name)
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(TerminalType.randgen_name, randgen_name)
            )

        if self._text.startswith(RAND_GEN_PERCENT_SYM, self._next_index):
            self._next_index += 1
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(
                    TerminalType.percentgen_marker, RAND_GEN_PERCENT_SYM
                )
            )

            if not self._try_to_match_rule(RulePercentGen):
                self.error_msg += \
                    " Percentage for the random generation is required after " + \
                    "its marker character ('" + RAND_GEN_PERCENT_SYM + "')."
                return False

        return True



"""
Module `chatette.cli.interactive_commands.hide_command`.
Contains the strategy class that represents the interactive mode command
`hide` which hides unit definitions (storing them somewhere to be able to
unhide them later).
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.units.ast import AST


class HideCommand(CommandStrategy):
    usage_str = 'hide <unit-type> "<unit-name>"'
    stored_units = {"alias": dict(), "slot": dict(), "intent": dict()}
    stored_variations = {"alias": dict(), "slot": dict(), "intent": dict()}
    def __init__(self, command_str, quiet=False):
        super(HideCommand, self).__init__(command_str, quiet)
        self._units_to_delete = []
        self._var_to_delete = []

    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        try:
            unit = AST.get_or_create()[unit_type][unit_name]
            if variation_name is None:
                self.stored_units[unit_type.name][unit_name] = unit
                self._units_to_delete.append((unit_type, unit_name))
                self.print_wrapper.write(
                    unit_type.name.capitalize() + " '" + unit_name + \
                    "' was successfully hidden."
                )
            else:
                if variation_name not in unit._variation_rules:
                    self.print_wrapper.error_log(
                        "Couldn't find variation '" + variation_name + \
                        "' in " + unit_type.name + " '" + unit_name + "'."
                    )
                    return
                self._var_to_delete.append((unit_type, unit_name, variation_name))
                rules = unit._variation_rules[variation_name]
                if unit_name not in self.stored_variations[unit_type.name]:
                    self.stored_variations[unit_type.name][unit_name] = \
                        {variation_name: rules}
                else:
                    self.stored_variations[unit_type.name][unit_name][variation_name] = \
                        rules
                self.print_wrapper.write(
                    "Variation '" + variation_name + "' of " + \
                    unit_type.name + " '" + unit_name + \
                    "' was successfully hidden."
                )
        except KeyError:
            self.print_wrapper.write(
                unit_type.name.capitalize() + " '" + unit_name + \
                "' was not defined."
            )

    def finish_execution(self):
        for (unit_type, unit_name) in self._units_to_delete:
            AST.get_or_create().delete_unit(unit_type, unit_name)
        self._units_to_delete = []

        for (unit_type, unit_name, variation_name) in self._var_to_delete:
            AST.get_or_create().delete_unit(
                unit_type, unit_name, variation_name
            )
        self._var_to_delete = []



# coding: utf-8
"""
Module `chatette.modifiers.casegen`
Contains the different functions that allow to apply
the case generation modifier to one or several examples.
"""

from random import random
from copy import deepcopy


def modify_nb_possibilities(unmodified_nb_possibilities):
    """
    Returns the number of possibilities of generation for an item that has
    a case generation modifier, given the number of possibilities for
    the same item without this modifier.
    """
    return 2 * unmodified_nb_possibilities


def modify_example(example):
    """
    Modifies the generated example `example` by applying
    the case generation modifier.
    Returns the modified example.
    """
    if random() < 0.5:
        return with_leading_upper(example)
    return with_leading_lower(example)


def make_all_possibilities(examples):
    """
    Given the list of examples `examples`, constructs and returns a list
    of all possible examples after the case generation modifier applied.
    """
    result = []
    for ex in examples:
        lowercase_ex = with_leading_lower(ex)
        result.append(lowercase_ex)
        uppercase_ex = with_leading_upper(deepcopy(ex))
        if not uppercase_ex.is_dup(lowercase_ex):
            result.append(uppercase_ex)
    return result

############# Utility functions ##############
def may_change_leading_case(text):
    """
    Checks whether the string `text` can
    change the letter case of its leading letter.
    """
    for c in text:
        if c.isalpha():
            return True
        if c.isspace():
            continue
        return False
    return False


def with_leading_upper(example):
    """
    Changes the leading letter of the text of example `example` to uppercase.
    """
    text = example.text
    for (i, c) in enumerate(text):
        if not c.isspace():
            text = text[:i] + text[i].upper() + text[(i + 1):]
            break
    example.text = text
    return example

def with_leading_lower(example):
    """
    Changes the leading letter of the text of example `example` to uppercase.
    """
    text = example.text
    for (i, c) in enumerate(text):
        if not c.isspace():
            text = text[:i] + text[i].lower() + text[(i + 1):]
            break
    example.text = text
    return example



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_content_rule_and_choice`
Contains the definition of the class that represents the lexing rule
to tokenize a line that is the contents of a rule inside a unit definition,
and the definition of the class that represents the lexing rule
to tokenize a choice (inside the contents of a rule).
Both classes are defined here to prevent circular imports (cf. below).
"""

from chatette.deprecations import Deprecations
from chatette.parsing.input_file_manager import InputFileManager

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import \
    CHOICE_START, CHOICE_END, CHOICE_SEP, \
    OLD_CHOICE_START, OLD_CHOICE_END, OLD_CHOICE_SEP, \
    CASE_GEN_SYM

from chatette.parsing.lexing.rule_word import RuleWord
from chatette.parsing.lexing.rule_unit_ref import RuleUnitRef
from chatette.parsing.lexing.rule_arg_assignment import \
    RuleArgAssignment
from chatette.parsing.lexing.rule_whitespaces import RuleWhitespaces
from chatette.parsing.lexing.rule_rand_gen import RuleRandGen


class RuleContentRule(LexingRule):
    def _apply_strategy(self, **kwargs):
        if self._match_one_of(
            [RuleWord, RuleChoice, RuleUnitRef, RuleArgAssignment],
            self._next_index,
            **kwargs
        ):
            if not self._try_to_match_rule(RuleWhitespaces):
                self.error_msg = None
            return True

        return False

# NOTE Required to put it here rather than in its own module
#      to prevent circular imports.
#      Indeed, instances of `RuleContentRule` can contain instances of
#      `RuleChoice`, which in turn can contain instances of `RuleContentRule`.
class RuleChoice(LexingRule):
    def _apply_strategy(self, **kwargs):
        start_char = None
        if self._text.startswith(OLD_CHOICE_START, self._next_index):
            start_char = OLD_CHOICE_START
            sep_char = OLD_CHOICE_SEP
            end_char = OLD_CHOICE_END
            Deprecations.get_or_create().warn_old_choice(
                *(InputFileManager \
                    .get_or_create() \
                    .get_current_line_information())
            )
        elif self._text.startswith(CHOICE_START, self._next_index):
            start_char = CHOICE_START
            sep_char = CHOICE_SEP
            end_char = CHOICE_END

        if start_char is None:
            self.error_msg = \
                "Invalid token. Expected a choice to start there (starting " + \
                "with '" + CHOICE_START + "' or '" + OLD_CHOICE_START + "')."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(
            LexicalToken(TerminalType.choice_start, start_char)
        )

        if self._text.startswith(CASE_GEN_SYM, self._next_index):
            self._next_index += 1
            self._update_furthest_matched_index()
            self._tokens.append(
                LexicalToken(TerminalType.casegen_marker, CASE_GEN_SYM)
            )

        if not self._try_to_match_rule(RuleWhitespaces):
            self.error_msg = None

        while True:
            if self._text.startswith(sep_char, self._next_index):
                self._next_index += 1
                self._update_furthest_matched_index()
                self._tokens.append(
                    LexicalToken(TerminalType.choice_sep, sep_char)
                )
                if not self._try_to_match_rule(RuleWhitespaces):
                    self.error_msg = None

            rule_content_rule = RuleContentRule(self._text, self._next_index)
            if not rule_content_rule.matches(inside_choice=True):
                self.error_msg = None
                self._update_furthest_matched_index(rule_content_rule)
                break
            self._next_index = rule_content_rule.get_next_index_to_match()
            self._update_furthest_matched_index(rule_content_rule)
            self._tokens.extend(rule_content_rule.get_lexical_tokens())

        if not self._try_to_match_rule(RuleRandGen):
            self.error_msg = None

        if not self._text.startswith(end_char, self._next_index):
            self.error_msg = \
                "Invalid token. Unmatched choice opening character. " + \
                "Expected the choice to end here (using " + \
                "character '" + end_char + "')."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(LexicalToken(TerminalType.choice_end, end_char))

        return True



"""
Module `chatette.terminal_writer`.
Contains a wrapper of the output of commands, that can write to the terminal
(stdout) or to a file.
"""

from __future__ import print_function
import io
import os.path
from enum import Enum


class RedirectionType(Enum):  # QUESTION is it possible to merge this with relevant strings?
    truncate = 1
    append = 2
    quiet = 3


class TerminalWriter(object):
    """Wrapper of `print` that can write to stdout or to a file."""
    def __init__(self, redirection_type=RedirectionType.append,
                 redirection_file_path=None):
        self.redirection_file_path = redirection_file_path
        self.buffered_text = None

        self._file_mode = None
        self.set_redirection_type(redirection_type)

    def reset(self):
        self.redirection_file_path = None
        self.buffered_text = None
    def set_redirection_type(self, redirection_type):
        """
        Sets redirection type.
        @pre: `redirection_type` is of type `RedirectionType`.
        """
        if redirection_type == RedirectionType.append:
            self._file_mode = 'a+'
        elif redirection_type == RedirectionType.truncate:
            self._file_mode = 'w+'
        elif redirection_type == RedirectionType.quiet:
            self._file_mode = 'quiet'

    def get_redirection(self):
        """
        Returns a 2-tuple containing the type and file path of the redirection.
        If this wrapper doesn't redirect to any file (or ignore prints),
        returns `None`.
        """
        if self._file_mode is None:
            return None
        if self._file_mode == 'quiet':
            return (RedirectionType.quiet, None)
        if self._file_mode == 'a+':
            return (RedirectionType.append, self.redirection_file_path)
        if self._file_mode == 'w+':
            return (RedirectionType.truncate, self.redirection_file_path)
        return None


    def write(self, text):
        if self.redirection_file_path is None and self._file_mode is None:
            print(text)
        elif self._file_mode == 'quiet':
            return
        else:
            if self.buffered_text is None:
                self.buffered_text = str(text)
            else:
                self.buffered_text += '\n' + str(text)

    def error_log(self, text):
        processed_text = ''.join(['\t' + line + '\n'
                                  for line in text.split('\n')])
        self.write("[ERROR]"+processed_text[:-1])


    def flush(self):
        """
        Flushes the buffered text to the redirection file
        if such a file is provided.
        """
        if self.redirection_file_path is not None:
            # Create file if it doesn't exist
            if not os.path.isfile(self.redirection_file_path):
                io.open(self.redirection_file_path, 'w+').close()
            # Write to the file if needed
            if self.buffered_text is not None:
                with io.open(self.redirection_file_path, self._file_mode) as f:
                    print(self.buffered_text, '\n', sep='', file=f)
        self.buffered_text = None



# coding: utf-8
"""
Module `chatette.configuration`
Contains the singleton storing the current configuration
of the program.
"""

from chatette.utils import Singleton
from chatette.log import print_warn


class Configuration(Singleton):
    """
    Singleton containing the current configuration of
    the whole program.
    """
    _instance = None
    def __init__(self):
        self.caching_level = 100  # out of 100

    def set_caching_level(self, new_level):
        print_warn(
            "Setting caching level to " + str(new_level) + \
            " for performance reasons."
        )
        if new_level < 0 or new_level > 100:
            raise ValueError(
                "Tried to set the caching level to an invalid level (" + \
                str(new_level) + ")."
            )
        self.caching_level = new_level



"""
Module `chatette.cli.interactive_command.save_command`.
Contains the strategy class that represents the interactive mode command
`save` which writes a template file that, when parsed, would make a parser
that is in the state of the current parser.
"""

from __future__ import print_function
import io

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.utils import cast_to_unicode

from chatette.units.ast import AST


class SaveCommand(CommandStrategy):
    usage_str = 'save <template-file-path>'

    def execute(self):
        if len(self.command_tokens) < 2:
            self.print_wrapper.error_log("Missing some arguments\nUsage: " +
                                         self.usage_str)
            return

        template_filepath = self.command_tokens[1]
        definitions = AST.get_or_create()
        with io.open(template_filepath, 'w+') as f:
            for intent_name in definitions._intent_definitions:
                intent = definitions._intent_definitions[intent_name]
                print(cast_to_unicode(intent.as_template_str() + '\n'), file=f)
            print(cast_to_unicode(''), file=f)
            for alias_name in definitions._alias_definitions:
                alias = definitions._alias_definitions[alias_name]
                print(cast_to_unicode(alias.as_template_str() + '\n'), file=f)
            print(cast_to_unicode(''), file=f)
            for slot_name in definitions._slot_definitions:
                slot = definitions._slot_definitions[slot_name]
                print(cast_to_unicode(slot.as_template_str() + '\n'), file=f)
            print(cast_to_unicode(''), file=f)
        self.print_wrapper.write("Template file successfully written.")


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



"""
Module `chatette.cli.interactive_commands.parse_command`.
Contains the strategy class that represents the interactive mode command
`parse` which parses a new template file.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.facade import Facade


class ParseCommand(CommandStrategy):

    def execute(self):
        """
        Implements the command `parse`,
        parsing a new template file using the current parser.
        """
        if len(self.command_tokens) <= 1:
            self.print_wrapper.error_log("Missing template file path\nUsage: " +
                                         "'parse <file_path>'")
            return
        file_path = self.command_tokens[1]

        if Facade.was_instantiated():
            facade = Facade.get_or_create()
            facade.parse_file(file_path)
        else:
            facade = Facade(file_path)
            facade.run_parsing()


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_file_inclusion`
Contains the definition of the class that represents the lexing rule
that has to do with a line that includes a file.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import FILE_INCLUSION_SYM, find_next_comment


class RuleFileInclusion(LexingRule):
    def _apply_strategy(self, **kwargs):
        if not self._text.startswith(FILE_INCLUSION_SYM, self._next_index):
            self.error_msg = \
                "Invalid token. Expected a file to be included there " + \
                "(starting with '" + FILE_INCLUSION_SYM + "')."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()
        self._tokens.append(
            LexicalToken(
                TerminalType.file_inclusion_marker, FILE_INCLUSION_SYM
            )
        )

        if self._text[self._next_index].isspace():
            self.error_msg = \
                "Invalid token. Expected a file path here, got a whitespace."
            return False

        comment_start = find_next_comment(self._text, self._next_index)
        if comment_start is not None:
            file_path = self._text[self._next_index:comment_start].rstrip()
        else:
            file_path = self._text[self._next_index:].rstrip()

        self._next_index += len(file_path)
        self._update_furthest_matched_index()
        self._tokens.append(LexicalToken(TerminalType.file_path, file_path))

        return True



"""
Module `chatette.cli.interactive_commands.definition_command`.
Contains the strategy class that represents the interactive mode command
`examples` which generates several (or all) possible examples for a given unit.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.units.ast import AST


class ExamplesCommand(CommandStrategy):
    usage_str = 'examples <unit-type> "<unit-name>" [<number-examples>]'
    def __init__(self, command_str, quiet=False):
        super(ExamplesCommand, self).__init__(command_str, quiet)
        self.nb_examples = None

    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        if self.nb_examples is None:  # Caching the parsed number
            self.nb_examples = -1  # All examples should be generated
            if len(self.command_tokens) > 3:
                try:
                    self.nb_examples = int(self.command_tokens[3])
                except ValueError:
                    self.print_wrapper.error_log(
                        "The number of examples to be generated is invalid: " + \
                        "it must be an integer (no other characters allowed)."
                    )
                    return
        try:
            definition = \
                AST.get_or_create()[unit_type][unit_name]
        except KeyError:
            self.print_wrapper.write(
                unit_type.name.capitalize() + " '" + unit_name + \
                "' is not defined."
            )
            return
        if self.nb_examples != -1:
            examples = definition.generate_nb_possibilities(
                self.nb_examples, variation_name=variation_name
            )
        else:
            examples = definition.generate_all(variation_name=variation_name)
        self.print_wrapper.write(
            "Examples for " + unit_type.name + " '" + unit_name + "':"
        )
        for ex in examples:
            self.print_wrapper.write(ex)
        self.print_wrapper.write("")

    def finish_execution(self):
        self.nb_examples = None



# coding: utf-8
"""
Module `chatette.parsing.lexing.rule_unit_start`
Contains the class that represents a lexing rule to tokenize
the start of a unit definition or reference.
"""

from chatette.parsing.lexing.lexing_rule import LexingRule
from chatette.parsing.lexing import LexicalToken, TerminalType
from chatette.parsing.utils import \
    ALIAS_SYM, SLOT_SYM, INTENT_SYM, UNIT_START_SYM


class RuleUnitStart(LexingRule):
    def _apply_strategy(self, **kwargs):
        """
        `kwargs` can contain a value with key `extracting_decl`.
        This is a boolean that should be `True` iff the rule should consider
        it is parsing a unit declaration and `False` if it is parsing
        a unit reference.
        If `kwargs` doesn't contain `extracting_decl`, defaults to `True`.
        """
        extracting_decl = kwargs.get("extracting_decl", True)
        if self._text.startswith(ALIAS_SYM, self._next_index):
            if extracting_decl:
                terminal_type = TerminalType.alias_decl_start
            else:
                terminal_type = TerminalType.alias_ref_start
            text_start = ALIAS_SYM
        elif self._text.startswith(SLOT_SYM, self._next_index):
            if extracting_decl:
                terminal_type = TerminalType.slot_decl_start
            else:
                terminal_type = TerminalType.slot_ref_start
            text_start = SLOT_SYM
        elif self._text.startswith(INTENT_SYM, self._next_index):
            if extracting_decl:
                terminal_type = TerminalType.intent_decl_start
            else:
                terminal_type = TerminalType.intent_ref_start
            text_start = INTENT_SYM
        else:
            self.error_msg = \
                "Invalid token. Expected a unit start here (starting with " + \
                "either '" + ALIAS_SYM + "', '" + SLOT_SYM + "' or '" + \
                INTENT_SYM + "'."
            return False
        self._next_index += 1
        self._update_furthest_matched_index()

        if self._text.startswith(UNIT_START_SYM, self._next_index):
            self._tokens.append(
                LexicalToken(terminal_type, text_start + UNIT_START_SYM)
            )
            self._next_index += 1
            self._update_furthest_matched_index()
            return True

        self.error_msg = \
            "Invalid token. Expected a start of unit here (starting with '" + \
            UNIT_START_SYM + "'). Did you mean to escape the previous '" + \
            text_start + '?'
        return False



"""
Module `chatette.cli.interactive_commands.stats_command`.
Contains the strategy class that represents the interactive mode command
`stats` which shows statistics about the parsing.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy
from chatette.statistics import Stats


class StatsCommand(CommandStrategy):

    def execute(self):
        """Implements the command `stats`, printing parsing statistics."""
        self.print_wrapper.write(str(Stats.get_or_create()))


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



"""
Module `chatette.cli.interactive_commands.execute_command`.
Contains the strategy class that represents the interactive mode command
`execute` which opens a file containing a bunch of commands and executes them.
"""

import io

from chatette.cli.interactive_commands.command_strategy import \
    CommandStrategy, REDIRECTION_SYM, REDIRECTION_APPEND_SYM


class ExecuteCommand(CommandStrategy):
    usage_str = 'execute <command-filepath>'

    def execute(self):
        """
        Implements the command `execute` which executes each line of a file
        as if they were commands (unless they start with a double slash `//`).
        """
        if len(self.command_tokens) < 2:
            self.print_wrapper.error_log(
                "Missing some arguments\nUsage: " + self.usage_str
            )
            return

        # TODO there may be a better way to get the file name
        with io.open(self.split_exact_unit_name(self.command_tokens[1])[0], 'r') as command_file:
            commands = [
                line.rstrip()
                for line in command_file.readlines()
                if not line.lstrip().startswith("//")
            ]
        redirection_tuple = self.print_wrapper.get_redirection()
        if redirection_tuple is not None:
            (_, redirection_filepath) = redirection_tuple
            if redirection_filepath is not None:
                text_to_append = \
                    ' ' + REDIRECTION_APPEND_SYM + ' ' + redirection_filepath
                for (i, cmd) in enumerate(commands):
                    if (
                        REDIRECTION_APPEND_SYM not in cmd
                        and REDIRECTION_SYM not in cmd
                    ):
                        commands[i] += text_to_append
        return commands


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



# coding: utf-8
"""
Module `chatette.units.modifiable`
Contains all the classes representing items
whose generation can be modified by modifiers.
All those classes extend the abstract class `ModifiableItem`
which is a sub-class of `GeneratingItem`.
"""

from random import choice as random_choice, uniform
from copy import deepcopy

from chatette.units.generating_item import GeneratingItem
from chatette.units import Example

from chatette.modifiers import casegen, argument, randgen


class ModifiableItem(GeneratingItem):
    def __init__(self, name, leading_space, modifiers):
        super(ModifiableItem, self).__init__(name, leading_space)
        if modifiers is None:
            raise ValueError(
                "Tried to instantiate a modifiable item " + \
                "with no modifier representation (" + \
                self.__class__.__name__ + ")"
            )
        self._modifiers_repr = modifiers
        # TODO add a check for modifiers


    def _make_empty_example(self):
        """
        Returns an example without any text or entity.
        Needed to be able to return a different type of example
        within intent definitions and other items.
        """
        return Example()


    def get_max_nb_possibilities(self, **kwargs):
        """
        Overriding.
        Calls the abstract method that computes the number of possibilities
        (without modifiers) and applies the modifiers.
        Caches the number of possibilities after computation.
        `kwargs` can contain `variation_name`.
        """
        if self._total_nb_possibilities is None:
            basic_nb_possibilities = self._compute_nb_possibilities(**kwargs)
            self._total_nb_possibilities = \
                self._modify_nb_possibilities(basic_nb_possibilities)
        return self._total_nb_possibilities


    # TODO this is quite hacky to avoid code duplication in subclasses (and not use the decorator pattern to avoid having too many objects)
    def generate_random(self, **kwargs):
        """
        Overriding.
        `kwargs` can contain the random name mapping `randgen_mapping` or
        `variation_name`.
        """
        variation_name = kwargs.get("variation_name", None)

        randgen_mapping = kwargs.get("randgen_mapping", None)
        if not self._should_generate(randgen_mapping):
            return self._make_empty_example()

        if variation_name is not None:
            max_nb_possibilities = \
                self.get_max_nb_possibilities(variation_name=variation_name)
        else:
            max_nb_possibilities = self.get_max_nb_possibilities()
        if isinstance(self._cached_examples, list):
            if (
                len(self._cached_examples) > 0
                and uniform(0, 1) <= \
                float(len(self._cached_examples)) / float(max_nb_possibilities)
            ):
                return random_choice(deepcopy(self._cached_examples))
        else:
            pass  # TODO dict case for unit definitions with variations

        if variation_name is not None:
            basic_example = \
                self._generate_random_strategy(variation_name=variation_name)
        else:
            basic_example = self._generate_random_strategy()
        if self._leading_space:
            basic_example.prepend(' ')
        return self._apply_modifiers(basic_example)


    # TODO this is quite hacky to avoid code duplication in subclasses (and not use the decorator pattern to avoid having too many objects)
    def generate_all(self, **kwargs):
        """Overriding."""
        if len(self._cached_examples) < self.get_max_nb_possibilities():
            basic_examples = self._generate_all_strategy(**kwargs)
            if self._leading_space:
                for ex in basic_examples:
                    ex.prepend(' ')
            all_examples = self._apply_modifiers_to_all(basic_examples)
            if self.get_max_cache_size() > 0:
                self._cached_examples = deepcopy(all_examples)
            return all_examples
        return deepcopy(self._cached_examples)


    def _modify_nb_possibilities(self, nb_possibilities):
        """
        Returns the number of possible different examples after application
        of the modifiers. `nb_possibilities` is the number of possibilities
        before the application of modifiers.
        """
        if self._modifiers_repr.casegen:
            nb_possibilities = \
                casegen.modify_nb_possibilities(nb_possibilities)
        if self._modifiers_repr.argument_value is not None:
            nb_possibilities = \
                argument.modify_nb_possibilities(nb_possibilities)
        if self._modifiers_repr.randgen:
            nb_possibilities = \
                randgen.modify_nb_possibilities(nb_possibilities)
        return nb_possibilities


    def _should_generate(self, randgen_mapping):
        """
        Returns `True` iff the current object should generate one example
        given its pre-modifiers (namely, the case generation modifier).
        `randgen_mapping` is the mapping from random generation modifier names
        to their boolean value.
        """
        if self._modifiers_repr.randgen:
            return \
                randgen.should_generate(
                    self._modifiers_repr.randgen.name,
                    self._modifiers_repr.randgen.percentage,
                    self._modifiers_repr.randgen.opposite,
                    randgen_mapping
                )
        return True


    def _apply_modifiers(self, example):
        """
        Returns the modified `example`
        after its post-modifiers have been applied.
        """
        if self._modifiers_repr.casegen:
            example = casegen.modify_example(example)
        if self._modifiers_repr.argument_value is not None:
            example = \
                argument.modify_example(
                    example, self._modifiers_repr.argument_value
                )
        return example

    def _apply_modifiers_to_all(self, examples):
        """
        Returns the list of examples `examples` with some additional examples,
        some removed examples and some modified examples as per application
        of its post-modifiers.
        """
        if self._modifiers_repr.casegen:
            examples = casegen.make_all_possibilities(examples)
        if self._modifiers_repr.argument_value is not None:
            examples = \
                argument.make_all_possibilities(
                    examples, self._modifiers_repr.argument_value
                )
        if self._modifiers_repr.randgen:
            examples = \
                randgen.make_all_possibilities(
                    examples, self._make_empty_example(),
                    self._modifiers_repr.randgen.name,
                    self._modifiers_repr.randgen.opposite
                )
        return examples


    def set_arg_name(self, new_arg_name):
        """
        Changes the name of the argument modifier and uncaches everything.
        """
        self._modifiers_repr.argument_name = new_arg_name
        self._reset_caches()
    def set_arg_value(self, new_arg_value):
        """
        Changes the value of the argument modifier and uncaches everything.
        """
        self._modifiers_repr.argument_value = new_arg_value
        self._reset_caches()
    def set_casegen(self, new_casegen=True):
        """
        Changes the value of the casegen modifier and uncaches everything.
        """
        self._modifiers_repr.casegen = new_casegen
        self._reset_caches()
    def set_randgen(self, new_randgen=True):
        """
        Changes the value of the randgen modifier and uncaches everything.
        """
        self._modifiers_repr.randgen = new_randgen
        self._reset_caches()
    def set_randgen_name(self, new_randgen_name):
        """
        Changes the value of the randgen name modifier and uncaches everything.
        """
        self._modifiers_repr.randgen_name = new_randgen_name
        self._reset_caches()
    def set_randgen_percent(self, new_randgen_percent=50):
        """
        Changes the value of the randgen name modifier and uncaches everything.
        """
        self._modifiers_repr.randgen_percent = new_randgen_percent



# coding: utf-8
"""
Module `chatette.units`
Contains every classes that represent what will be in the Abstract Syntax Tree
(AST) and the rules.
"""

from math import floor


class Example(object):
    """
    Represents an utterance (i.e. an example of an intent)
    that will later on be written in the output file(s).
    """
    def __init__(self, text=None, entities=None):
        if entities is None:
            entities = []
        if text is None:
            text = ""

        self.text = text
        self.entities = entities
        self._slot_value = None  # HACK used by slot to prevent code duplication

    def __repr__(self):
        # return "<'" + self.text + "' " + str(self.entities) + '>'
        return str(self)
    def __str__(self):
        return \
            "Text: '" + self.text + \
            "'\n\tEntities: " + str(self.entities)

    def as_dict(self):
        result = {"text": self.text, "entities": []}
        if self._slot_value is not None:
            result["slot-value"] = self._slot_value
        for entity in self.entities:
            result["entities"].append(entity.as_dict())
        return result

    def __hash__(self):
        entities_hash = 0
        for entity in self.entities:
            entities_hash += hash(entity)
        return hash(self.text) * 10000 + entities_hash  # QUESTION not sure this a very good hash

    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    def __ne__(self, other):
        return not self.__eq__(other)
    def is_dup(self, other):
        """
        Returns `True` if this example is a duplicate of `other`.
        A duplicate is an example with the same text.
        """
        return self.text == other.text

    def prepend(self, text_to_prepend):
        """
        Prepends `text_to_prepend` to the example
        and manages the pointer of the entities.
        """
        length = len(text_to_prepend)
        if length == 0:
            return
        self.text = text_to_prepend + self.text
        for entity in self.entities:
            entity._start_index += length
    def append(self, example_to_append):
        """
        Appends the text of `example_to_append` to the example text
        and manages its pointers to entities if needed.
        @post: don't use `example_to_append` any longer.
        """
        starting_index = len(self.text)
        self.text += example_to_append.text
        for entity in example_to_append.entities:
            entity._start_index += starting_index
            self.entities.append(entity)

    def remove_leading_space(self):
        """Removes the leading space of `self.text` if there is one."""
        if len(self.text) > 0 and self.text[0].isspace():
            self.text = self.text[1:]
            for entity in self.entities:
                entity._remove_leading_space()


class IntentExample(Example):
    def __init__(self, intent_name, text=None, entities=None):
        super(IntentExample, self).__init__(text, entities)
        self.intent_name = intent_name
    @classmethod
    def from_example(cls, example, intent_name):
        return cls(intent_name, example.text, example.entities)

    def as_dict(self):
        result = super(IntentExample, self).as_dict()
        result["intent-name"] = self.intent_name
        return result

    def __repr__(self):
        return \
            "<intent: " + self.intent_name + \
            " '" + self.text + "' " + str(self.entities) + ">"
    def __str__(self):
        return \
            "Intent: '" + self.intent_name + \
            "'\n\tText: '" + self.text + \
            "'\n\tEntities: " + str(self.entities)

    def __eq__(self, other):
        if isinstance(other, Example) and not isinstance(other, IntentExample):
            raise TypeError("Tried to compare an Example and an IntentExample")
        return super(IntentExample, self).__eq__(other)

    def __hash__(self):
        example_hash = super(IntentExample, self).__hash__()
        return hash(self.intent_name) * 1000000000 + example_hash  # QUESTION not sure this a very good hash


class Entity(object):
    """
    Represents an entity as it will be contained in examples
    (instances of `Example`).
    """
    def __init__(self, name, length, value=None, start_index=0):
        self.slot_name = name  # name of the entity (not the associated text)
        self.value = value
        self._len = length
        self._start_index = start_index

    def _remove_leading_space(self):
        """
        Adapts the start index and length if needed, after a leading space was
        removed from the example text.
        @returns: `True` if the entity still makes sense.
        """
        if self._start_index == 0:
            self._len -= 1
            if self._len <= 0:
                return False
        else:
            self._start_index -= 1
        return True

    def as_dict(self):
        return {
            "slot-name": self.slot_name,
            "value": self.value,
            "start-index": self._start_index,
            "end-index": self._start_index + self._len
        }

    def __repr__(self):
        representation = "entity '" + self.slot_name + "'"
        if self.value is not None:
            representation += ":'" + self.value + "'"
        return representation
    def __str__(self):
        return \
            self.slot_name + "@" + str(self._start_index) + "\t=>\t" + str(self.value)

    def __hash__(self):
        return \
            hash(self.slot_name +"@" + str(self._start_index) + ":" + str(self.value))

    def __eq__(self, other):
        return self.__dict__ == other.__dict__
    def __ne__(self, other):
        return not self.__eq__(other)


def add_example_no_dup(example_list, new_example):
    """
    Adds `new_example` to the list of examples `example_list`,
    and then removes duplicates.
    An example is a duplicate of another if they have the same text. The one
    that is kept is the one with the largest amount of entities.
    @pre: - `example_list` must be a sorted list (with key == text).
          - `example_list` does not contain duplicates.
    @post: the returned list is sorted.
    """
    # TODO maybe using sets would be a better idea?
    # Find closest example
    lo = 0
    hi = len(example_list) - 1
    i = int(floor(hi/2))
    found = False
    while lo <= hi:
        current_text = example_list[i].text
        if current_text == new_example.text:
            found = True
            break
        elif current_text > new_example.text:
            hi = float(i - 1)
        else:  # current_text < new_example.text
            lo = float(i + 1)

        i = int(floor(abs(hi - lo)/2) + lo)

    # Add example if needed
    if not found:
        example_list.insert(i, new_example)
    # Replace example if needed
    elif len(example_list[i].entities) < len(new_example.entities):
        example_list[i] = new_example
    return example_list

def extend_no_dup(example_list, new_examples):
    """
    Adds each example in `new_examples`
    iff it has no duplicate in `example_list`
    @pre: - `example_list` must be a sorted list (with key == text).
          - `example_list` does not contain duplicates.
          - `new_examples` must be a sorted list (with key == text).
          - `new_examples` does not contain duplicates.
    @post: the returned list is sorted.
    """
    if len(new_examples) == 0:
        return example_list
    if len(example_list) == 0:
        return new_examples
    # TODO there might be a more optimized way to do that
    for ex in new_examples:
        example_list = add_example_no_dup(example_list, ex)
    return example_list


def sort_by_texts(example_list):
    """
    Returns a sorted list of examples (with key == text).
    @pre: - `example_list` contains no duplicates.
    """
    return sorted(example_list, key=lambda ex: ex.text)


# coding: utf-8
"""
Module `chatette.units.ast`
Contains the data structure holding the Abstract Syntax Tree generated
when parsing the template files.
NOTE: this is not exactly an AST as it is not a tree, but it has the same
      purpose as an AST in a compiler, i.e. an intermediate representation
      of the parsed information used to generate the output.
"""

from chatette.utils import \
    Singleton, UnitType, remove_duplicates, extend_list_in_dict
from chatette.statistics import Stats

from chatette.units.modifiable.definitions.alias import AliasDefinition
from chatette.units.modifiable.definitions.slot import SlotDefinition
from chatette.units.modifiable.definitions.intent import IntentDefinition


class AST(Singleton):
    _instance = None
    def __init__(self):
        self._alias_definitions = dict()
        self._slot_definitions = dict()
        self._intent_definitions = dict()

        self.stats = Stats.get_or_create()


    def _get_relevant_dict(self, unit_type):
        """Returns the dict that stores units of type `unit_type`."""
        if unit_type == UnitType.alias:
            return self._alias_definitions
        elif unit_type == UnitType.slot:
            return self._slot_definitions
        elif unit_type == UnitType.intent:
            return self._intent_definitions
        else:
            raise TypeError(
                "Tried to get a definition with wrong type " + \
                "(expected alias, slot or intent)"
            )
    def __getitem__(self, unit_type):
        """
        Returns the dictionary requested.
        `unit_type` can be either a str ("alias", "slot" or "intent")
        or a `UnitType`.
        @raises: - `KeyError` if `unit_type` is an invalid str.
                 - `ValueError` if `unit_type´ is neither a str or a `UnitType`.
        """
        if isinstance(unit_type, str):
            if unit_type == UnitType.alias.value:
                unit_type = UnitType.alias
            elif unit_type == UnitType.slot.value:
                unit_type = UnitType.slot
            elif unit_type == UnitType.intent.value:
                unit_type = UnitType.intent
            else:
                raise KeyError(
                    "Invalid key for the AST: '" + unit_type + "'."
                )
        elif not isinstance(unit_type, UnitType):
            raise TypeError(
                "Invalid type of key: " + unit_type.__class__.__name__ + "."
            )
        return self._get_relevant_dict(unit_type)


    def _add_unit(self, unit_type, unit):
        """
        Adds the intent definition `intent` in the relevant dict.
        @raises: `ValueError` if the intent was already defined.
        """
        self.stats.new_variation_unit_declared(unit_type)
        relevant_dict = self._get_relevant_dict(unit_type)
        if unit.identifier in relevant_dict:
            raise ValueError(
                "Tried to declare " + unit_type.value + " '" + \
                unit.identifier + "' twice."
            )
        relevant_dict[unit.identifier] = unit
        self.stats.new_unit_declared(unit_type)

    def add_alias(self, alias):
        """
        Adds the alias definition `alias` in the relevant dict.
        @raises: `ValueError` if the alias was already defined.
        """
        # NOTE there might be a better way to check that the alias is not already defined without needing to call `get_relevant_dict`
        self._add_unit(UnitType.alias, alias)

    def add_slot(self, slot):
        """
        Adds the slot definition `slot` in the relevant dict.
        @raises: `ValueError` if the slot was already defined.
        """
        self._add_unit(UnitType.slot, slot)

    def add_intent(self, intent):
        """
        Adds the intent definition `intent` in the relevant dict.
        @raises: `ValueError` if the intent was already defined.
        """
        self._add_unit(UnitType.intent, intent)

    def add_unit(self, unit, unit_type=None):
        """
        Adds the unit definition `unit` in the relevant dict.
        If `unit_type` is `None`, detects the type of the definition
        by itself.
        @raises: - `TypeError` if the unit type is of an invalid type.
                 - `ValueError` if the unit was already declared.
        """
        if unit_type is None:
            if isinstance(unit, AliasDefinition):
                unit_type = UnitType.alias
            elif isinstance(unit, SlotDefinition):
                unit_type = UnitType.slot
            elif isinstance(unit, IntentDefinition):
                unit_type = UnitType.intent
            else:
                raise TypeError(  # Should never happen
                    "Tried to add something else than a unit definition " + \
                    "to the AST."
                )
        self._add_unit(unit_type, unit)


    def rename_unit(self, unit_type, old_name, new_name):
        """
        Changes the name of the unit `old_name` to `new_name` if it exists.
        @raises: - `KeyError` if the unit of type `unit_type` and name
                   `unit_name` wasn't declared.
                 - `ValueError` if the unit with name `new_name` already
                   existed.
        """
        relevant_dict = self._get_relevant_dict(unit_type)
        if old_name not in relevant_dict:
            raise KeyError(
                "Tried to rename " + unit_type.name + " '" + old_name + \
                "', but it wasn't declared."
            )
        if new_name in relevant_dict:
            raise ValueError(
                "Tried to rename " + unit_type.name + " '" + old_name + \
                "' to '" + new_name + "', but this " + unit_type.name + \
                " already existed."
            )
        unit = relevant_dict[old_name]
        del relevant_dict[old_name]
        unit.set_identifier(new_name)
        relevant_dict[new_name] = unit

    def delete_unit(self, unit_type, unit_name, variation_name=None):
        """
        Deletes the declared unit `unit_name` of type `unit_type`.
        If `variation_name` is not `None`, only deletes this particular
        variation for this unit.
        @raises: - `KeyError` if the unit `unit_name` wasn't declared.
                 - `KeyError` if the variation `variation_name` doesn't exist.
        """
        relevant_dict = self._get_relevant_dict(unit_type)
        if unit_name not in relevant_dict:
            raise KeyError(
                "Tried to delete " + unit_type.name + " '" + unit_name + \
                "', but it wasn't declared."
            )
        if variation_name is None:
            self.stats.one_unit_removed(unit_type)
            self.stats.several_variation_units_removed(
                unit_type, relevant_dict[unit_name].get_number_variations()
            )
            del relevant_dict[unit_name]
        else:
            relevant_dict[unit_name].delete_variation(variation_name)
            self.stats.one_variation_unit_removed(unit_type)


    def get_entities_synonyms(self):  # TODO move that into AST
        """
        Makes a dict of all the synonyms of entities
        based on the slot value they are assigned.
        """
        synonyms = dict()
        for slot_definition in self._slot_definitions:
            current_synonyms_dict = \
                self._slot_definitions[slot_definition].get_synonyms_dict()
            for slot_value in current_synonyms_dict:
                extend_list_in_dict(
                    synonyms, slot_value, current_synonyms_dict[slot_value]
                )
        return remove_duplicates(synonyms)


    def print_DBG(self):
        print("Aliases (" + str(len(self._alias_definitions)) + "):")
        for alias_name in self._alias_definitions:
            self._alias_definitions[alias_name].print_DBG()
        print("Slots (" + str(len(self._slot_definitions)) + "):")
        for slot_name in self._slot_definitions:
            self._slot_definitions[slot_name].print_DBG()
        print("Intents (" + str(len(self._intent_definitions)) + "):")
        for intent_name in self._intent_definitions:
            self._intent_definitions[intent_name].print_DBG()
        print()



#!/usr/bin/env python3
# coding: utf-8
"""
Module `chatette.log`
Contains logging functions used throughout the project.
"""

import sys


# pylint: disable=invalid-name
def print_DBG(txt):
    """Prints debug information on stdout."""
    print("[DBG] " + txt)


def print_warn(txt):
    """Warns the user using stdout."""
    print("\n[WARN] " + txt + "\n", file=sys.stderr)



# coding: utf-8
"""
Module `chatette.units.modifiable.definitions.unit_definition`
Contains the abstract class that is base for all unit definitions.
"""

from random import choice
from copy import deepcopy

from chatette.utils import UnitType
from chatette.units.modifiable import ModifiableItem
from chatette.units import extend_no_dup

from chatette.parsing.utils import SLOT_VAL_FIRST_RULE
from chatette.parsing import utils as putils

from chatette.statistics import Stats


class UnitDefinition(ModifiableItem):
    """Abstract class base for all unit definitions."""
    unit_type = None
    def __init__(self, identifier, modifiers):
        super(UnitDefinition, self).__init__(
            identifier, False, modifiers  # NOTE `False` corresponds to `leading_space`
        )
        self.identifier = identifier
        self._all_rules = []
        self._variation_rules = dict()


    def __contains__(self, variation_name):
        """
        Returns `True` if `variation_name` is a variation
        that is declared for the current definition.
        """
        return (variation_name in self._variation_rules)


    def set_identifier(self, new_identifier):
        """
        Changes the identifier of the current unit and recomputes its full name.
        @pre: `new_identifier` wasn't already in use for another unit of the
              same type.
        """
        self.identifier = new_identifier
        self._name = new_identifier
        self.full_name = self._compute_full_name()


    def _compute_nb_possibilities(self, variation_name=None):
        if variation_name is None:
            relevant_rules = self._all_rules
        elif variation_name in self._variation_rules:
            relevant_rules = self._variation_rules[variation_name]
        else:
            raise SyntaxError(
                "Referenced variation '" + str(variation_name) + "' for " + \
                self.full_name + ", but this variation has no rules " + \
                "associated to it."
            )

        acc = 0
        for rule in relevant_rules:
            acc += rule.get_max_nb_possibilities()
        return acc


    def _check_rule_validity(self, rule):
        """Raises a `ValueError` if the rule `rule` is not valid."""
        if rule.slot_value is not None:
            raise ValueError(
                "One of the rules in " + self.full_name + " was not valid: " + \
                "it contained a slot value. The invalid rule is " + str(rule)
            )

    def add_rule(self, rule, variation_name=None):
        """
        Adds the rule `rule` to the list of rules.
        If `variation_name` is not `None`, adds the rule to the corresponding
        variation.
        @raises: `ValueError` if `rule` is not valid.
        """
        self._check_rule_validity(rule)
        if variation_name in self._variation_rules:
            self._variation_rules[variation_name].append(rule)
        else:
            self._variation_rules[variation_name] = [rule]
        self._all_rules.append(rule)
    def add_all_rules(self, rules, variation_name=None):
        """
        Adds each of the rules `rule` to the list of rules.
        If `variation_name` is not `None`, adds the rules to the corresponding
        variation.
        @raises: `ValueError` if one of the rules in `rules` is not valid.
        """
        rule = None
        for rule in rules:
            self._check_rule_validity(rule)
        if variation_name in self._variation_rules:
            self._variation_rules[variation_name].extend(rule)
        else:
            self._variation_rules[variation_name] = rules
        self._all_rules.extend(rules)

    def _recompute_all_rules(self):
        """
        Given all the rules in `self._variation_rules`,
        recomputes all the possible rules and put them in `self._all_rules`.
        """
        all_rules = []
        for rules in self._variation_rules:
            all_rules.extend(rules)
        self._all_rules = all_rules

    def remove_rule(self, index, variation_name=None):
        """Removes the rule at `index`th rule."""
        # TODO decide what to do with `variation_name`
        if index < 0 or index >= len(self._all_rules):
            raise ValueError("Tried to remove rule at invalid index.")
        del self._all_rules[index]


    def _choose_rule(self, variation_name=None):
        """
        Returns a rule at random from the list of rules for this definition.
        Returns `None` if there are no rules.
        If `variation_name` is not `None`, the rule chosen should come
        from the corresponding variation (if it exists).
        """
        if variation_name is None:
            if len(self._all_rules) == 0:
                return None
            return choice(self._all_rules)
        else:
            if (
                variation_name not in self._variation_rules
                or len(self._variation_rules[variation_name]) == 0
            ):
                return None
            return choice(self._variation_rules[variation_name])


    def has_variation(self, variation_name):
        """Returns `True` iff `variation_name` was declared for this unit."""
        return (variation_name in self._variation_rules)
    def get_number_variations(self):
        if None in self._variation_rules:
            return len(self._variation_rules) - 1
        return len(self._variation_rules)
    def delete_variation(self, variation_name):
        """
        Deletes the variation named `variation_name`.
        @raises: - `KeyError` if the variation doesn't exist.
        """
        if not self.has_variation(variation_name):
            raise KeyError(
                "Tried to delete variation '" + str(variation_name) + \
                "' of " + self.full_name + ", but it wasn't defined."
            )
        Stats.get_or_create().one_variation_unit_removed(self.unit_type)
        del self._variation_rules[variation_name]
        self._recompute_all_rules()


    def _generate_random_strategy(self, variation_name=None):
        rule = self._choose_rule(variation_name)
        if rule is None:
            if variation_name is None:
                raise SyntaxError(
                    self.full_name.capitalize() + " does not have any rule to " + \
                    "generate."
                )
            else:
                raise SyntaxError(
                    self.full_name.capitalize() + " does not have any rule " + \
                    "associated to variation '" + str(variation_name) + "'."
                )

        example = rule.generate_random()
        example.remove_leading_space()
        if self.unit_type == UnitType.slot:
            example._slot_value = rule.slot_value
            if (
                example._slot_value is not None
                and example._slot_value.strip() == SLOT_VAL_FIRST_RULE
            ):
                if len(rule._contents) == 0:
                    raise SyntaxError(
                        "The slot value given for one of the rules of " + \
                        self.full_name + " references the first token of " + \
                        "an empty rule."
                    )
                example._slot_value = rule._contents[0]._name
        return example


    def generate_all(self, variation_name=None):
        """Overriding to prevent caching of examples for just one variation."""
        if not self._variation_rules:
            return super(UnitDefinition, self).generate_all()

        if isinstance(self._cached_examples, list):
            self._cached_examples = dict()
        if variation_name not in self._cached_examples:
            basic_examples = \
                self._generate_all_strategy(variation_name=variation_name)
            if self._leading_space:
                for ex in basic_examples:
                    ex.prepend(' ')
            self._cached_examples[variation_name] = \
                self._apply_modifiers_to_all(basic_examples)
        return deepcopy(self._cached_examples[variation_name])


    def _generate_all_strategy(self, variation_name=None):
        if variation_name is None:
            relevant_rules = self._all_rules
        elif variation_name in self._variation_rules:
            relevant_rules = self._variation_rules[variation_name]
        else:
            raise SyntaxError(
                "Tried to generate examples for variation '" + \
                str(variation_name) + "' of " + self.full_name + \
                ", but this variation has no rules associated to it."
            )

        generated_examples = []
        for rule in relevant_rules:
            current_examples = rule.generate_all()
            if self.unit_type == UnitType.slot:
                for ex in current_examples:
                    ex._slot_value = rule.slot_value
                    if (
                        ex._slot_value is not None
                        and ex._slot_value.strip() == SLOT_VAL_FIRST_RULE
                    ):
                        if len(rule._contents) == 0:
                            raise SyntaxError(
                                "The slot value given for one of the rules " + \
                                "of " + self.full_name + \
                                " references the first token of an empty rule."
                            )
                        ex._slot_value = rule._contents[0]._name
            for ex in current_examples:
                ex.remove_leading_space()
            generated_examples = \
                extend_no_dup(generated_examples, current_examples)
        return generated_examples


    def short_description(self):
        """
        Returns a short description (as a `str`) that can be displayed to the
        user.
        """
        desc = self.full_name + "\n"
        desc += self._modifiers_repr.short_description()
        if None in self._variation_rules:
            desc += str(max(0, len(self._variation_rules) - 1)) + " variation(s)"
        else:
            desc += str(len(self._variation_rules)) + " variation(s)"
        if len(self._variation_rules) > 1:
            desc += ':'
        for (i, variation_name) in enumerate(self._variation_rules):
            if variation_name is None:
                continue
            desc += "\n\t- " + variation_name
            if i >= 12:
                desc += "\n\t- ..."
                break
        return desc

    def as_template_str(self):
        result = ""
        for variation_name in self._variation_rules:
            result += putils.get_template_unit_sym(self.unit_type)
            result += putils.UNIT_START_SYM
            result += putils.get_template_pre_modifiers(self._modifiers_repr)
            result += self._name
            result += putils.get_template_post_modifiers(self._modifiers_repr)
            if variation_name is not None:
                result += putils.VARIATION_SYM + variation_name
            result += putils.UNIT_END_SYM

            for rule in self._variation_rules[variation_name]:
                result += '\n\t' + rule.as_template_str()
        return result



# coding: utf-8
"""
Module `chatette.units.modifiable.definitions.slot`
Contains the class representing a slot definition.
"""

from chatette.utils import UnitType, append_to_list_in_dict, extend_list_in_dict
from chatette.units import Entity
from chatette.units.modifiable.definitions.unit_definition import \
    UnitDefinition


class SlotDefinition(UnitDefinition):
    """Represents an slot definition."""
    unit_type = UnitType.slot
    def __init__(self, *args, **kwargs):
        super(SlotDefinition, self).__init__(*args, **kwargs)
        self._synonyms = None

    def _compute_full_name(self):
        return "slot '" + self._name + "'"


    def _check_rule_validity(self, rule):
        """Override."""
        pass


    def _generate_random_strategy(self, variation_name=None):
        generated_example = \
            super(SlotDefinition, self)._generate_random_strategy(
                variation_name=variation_name
            )

        slot_value = generated_example._slot_value
        if slot_value is None:
            slot_value = generated_example.text
        generated_example.entities.append(
            Entity(self._name, len(generated_example.text), slot_value)
        )
        generated_example._slot_value = None

        return generated_example

    def _generate_all_strategy(self, variation_name=None):
        generated_examples = \
            super(SlotDefinition, self)._generate_all_strategy(
                variation_name=variation_name
            )

        for ex in generated_examples:
            slot_value = ex._slot_value
            if slot_value is None:
                slot_value = ex.text
            ex.entities.append(
                Entity(self._name, len(ex.text), slot_value)
            )
            ex._slot_value = None

        return generated_examples


    def get_synonyms_dict(self):
        if self._synonyms is None:
            self._synonyms = dict()
            for rule in self._all_rules:
                texts = [ex.text for ex in rule.generate_all()]
                if rule.slot_value is None:
                    for text in texts:
                        append_to_list_in_dict(self._synonyms, text, text)
                else:
                    extend_list_in_dict(self._synonyms, rule.slot_value, texts)
        return self._synonyms



# coding: utf-8
"""
Module `chatette.parsing.input_file_manager`
Contains the definition of the singleton `ÌnputFileManager`
which is in charge of managing the opening, closing and (read) accesses
to template files.
"""

import os.path

from chatette.utils import Singleton, cast_to_unicode
from chatette.parsing.line_count_file_wrapper import \
    LineCountFileWrapper
from chatette.statistics import Stats


class FileAlreadyOpened(ValueError):
    pass


class InputFileManager(Singleton):
    """
    Singleton in charge of managing the opening, closing and read accesses
    to templates file(s). In charge of raising the information about
    syntax errors if asked by other objects (namely the ones in charge of
    lexing and parsing the files).
    """
    _instance = None
    @classmethod
    def get_or_create(cls, file_path=None):
        """
        Returns the instance of the class (representing the singleton)
        or instantiate it if no instance already exists.
        If `file_path` is provided, starts to read it.
        """
        if cls._instance is None:
            cls._instance = cls(file_path)
        elif file_path is not None:
            cls._instance.open_file(file_path)
        return cls._instance

    def __init__(self, file_path=None):
        self._current_file = None
        self._opened_files = []

        self._last_read_line = None  # str

        if file_path is not None:
            self.open_file(file_path)

    def open_file(self, file_path):
        """
        Opens the file at `file_path` if and only if it wasn't opened before.
        Stores the currently read file for further reading if needed.
        `file_path` is understood with respect to the currently being parsed
        file (rather than working directory),
        unless it is an absolute path or there is no file currently being
        parsed.
        @raises: - `FileAlreadyOpened` if the file at `file_path`
                   was already opened.
        """
        file_path = cast_to_unicode(file_path)
        if not os.path.isabs(file_path):
            if self._current_file is None:
                file_path = cast_to_unicode(os.path.abspath(file_path))
            else:
                file_path = \
                    os.path.join(
                        cast_to_unicode(os.path.dirname(self._current_file.name)),
                        file_path
                    )

        opened_file_paths = [f.name for f in self._opened_files]
        if file_path in opened_file_paths:
            raise FileAlreadyOpened(
                "Tried to read file '" + file_path + "' several times."
            )

        if self._current_file is not None:
            self._opened_files.append(self._current_file)
        try:
            self._current_file = LineCountFileWrapper(file_path)
            Stats.get_or_create().new_file_parsed()
        except IOError as e:
            if len(self._opened_files) > 0:
                self._current_file = self._opened_files.pop()
            raise e

    def close_all_files(self):
        """Closes all the opened files."""
        for f in self._opened_files:
            if not f.closed:
                f.close()
        if self._current_file is not None and not self._current_file.closed:
            self._current_file.close()

    def _close_current_file(self):
        """
        Closes current file and
        continues reading the file previously being read (if any).
        """
        if self._current_file is not None:
            self._current_file.close()
        if len(self._opened_files) > 0:
            self._current_file = self._opened_files.pop()
        else:
            self._current_file = None


    def get_current_file_information(self):
        if self._current_file is None:
            return (None, None)
        return (self._current_file.name, self._current_file.line_nb)

    def get_current_file_name(self):
        if self._current_file is None:
            return None
        return self._current_file.name

    def get_current_line_information(self):
        if self._current_file is None:
            return (None, None, self._last_read_line)
        return (
            self._current_file.name, self._current_file.line_nb,
            self._last_read_line
        )


    def syntax_error(self, message, char_index=0, word_to_find=None):
        """
        Closes all files and raises a `SyntaxError`
        pointing at the index `char_index` in the last read line.
        """
        if word_to_find is not None and self._last_read_line is not None:
            char_index = self._last_read_line.find(word_to_find)

        self.close_all_files()
        if self._current_file is not None:
            raise SyntaxError(  # BUG the file name and line read are sometimes not displayed (only the line number)
                message,
                (
                    self._current_file.name,
                    self._current_file.line_nb, char_index,
                    self._last_read_line
                )
            )
        raise SyntaxError(
            message,
            (None, None, char_index, self._last_read_line)
        )


    def read_line(self):
        """
        Reads the next line of `self._current_file` and
        returns it without the trailing new line (`\n`)
        (and any trailing whitespaces).
        If the file was entirely read, closes it and
        continues to read the file that was previously being read (if any).
        Returns `None` if there is no file left to read.
        """
        # TODO make this a generator?
        line = self._current_file.readline()
        while line == '':  # EOF
            self._close_current_file()
            if self._current_file is None:  # No more files to read
                return None
            line = self._current_file.readline()
        line = line.rstrip()
        self._last_read_line = line
        return line



# coding: utf-8
import io
import json
import os

from chatette.utils import cast_to_unicode
from ._base import Adapter


class JsonListAdapter(Adapter):
    @classmethod
    def _get_file_extension(cls):
        return "jsonl"

    def prepare_example(self, example):
        return json.dumps(
            cast_to_unicode(example.as_dict()),
            ensure_ascii=False, sort_keys=True
        )

    def _write_batch(self, output_file_handle, batch):
        output_file_handle.writelines([
            self.prepare_example(example) + "\n"
            for example in batch.examples
        ])

    def write(self, output_directory, examples, synonyms):
        super(JsonListAdapter, self).write(output_directory, examples, synonyms)

        processed_synonyms = self.__format_synonyms(synonyms)
        if processed_synonyms is not None:
            synonyms_file_path = os.path.join(output_directory, "synonyms.json")
            with io.open(synonyms_file_path, 'w') as output_file:
                output_file.write(
                    json.dumps(
                        cast_to_unicode(processed_synonyms),
                        ensure_ascii=False, sort_keys=True, indent=2
                    )
                )


    @classmethod
    def __format_synonyms(cls, synonyms):
        result = {
            key: values
            for (key, values) in synonyms.items()
            if len(values) > 1 or values[0] != key
        }
        if not result:
            return None
        return result



"""
Module `chatette.cli.interactive_commands.rule_command`.
Contains the strategy class that represents the interactive mode command
`rule` which generates as many examples as asked that the provided rule
can generate.
"""

from chatette.cli.interactive_commands.command_strategy import CommandStrategy

from chatette.parsing.parser import Parser


class RuleCommand(CommandStrategy):
    def execute(self):
        """
        Implements the command `rule` which generates a certain number of
        examples according to a provided rule.
        """
        if len(self.command_tokens) < 2:
            self.print_wrapper.error_log("Missing some arguments\nUsage: " +
                                         'rule "<rule>" [<number-of-examples]')
            return

        rule_str = CommandStrategy.remove_quotes(self.command_tokens[1])
        nb_examples = None
        if len(self.command_tokens) >= 3:
            try:
                nb_examples = int(self.command_tokens[2])
            except ValueError:
                self.print_wrapper.error_log(
                    "The number of examples asked (" + \
                    self.command_tokens[2] + ") is a valid integer."
                )

        parser = Parser(None)
        rule_tokens = parser.lexer.lex("\t" + rule_str)
        # pylint: disable=protected-access
        rule = parser._parse_rule(rule_tokens[1:])

        if nb_examples is None:
            examples = rule.generate_all()
        else:
            examples = rule.generate_nb_possibilities(nb_examples)
        self.print_wrapper.write("Generated examples:")
        for ex in examples:
            self.print_wrapper.write(str(ex))


    # Override abstract methods
    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        raise NotImplementedError()
    def finish_execution(self):
        raise NotImplementedError()



# coding: utf-8
"""
Module `chatette.parsing.lexing`
Contains every class that has to do with lexing
(transforming a text in a sequence of labelled tokens).
"""

from enum import Enum
from chatette.parsing.utils import ESCAPEMENT_SYM, ESCAPABLE_CHARS


# Supported tokens
class TerminalType(Enum):
    """Enum of terminals types that will be used by the lexer."""
    # Special
    ignore = 0
    indentation = 1
    whitespace = 2
    comment = 3
    # File inclusion
    file_inclusion_marker = 4
    file_path = 5
    # Unit declaration
    alias_decl_start = 6
    alias_decl_end = 7
    slot_decl_start = 8
    slot_decl_end = 9
    intent_decl_start = 10
    intent_decl_end = 11
    unit_identifier = 12
    # Annotations
    annotation_start = 13
    annotation_end = 14
    separator = 15
    key = 16
    value = 17
    key_value_connector = 18
    encloser = 19
    # Rule contents
    word = 20
    choice_start = 21
    choice_end = 22
    choice_sep = 23
    alias_ref_start = 24
    alias_ref_end = 25
    slot_ref_start = 26
    slot_ref_end = 27
    intent_ref_start = 28
    intent_ref_end = 29
    slot_val_marker = 30  # '='
    slot_val = 31
    # Modifiers
    casegen_marker = 32
    arg_marker = 33
    arg_name = 34
    arg_value = 35
    arg_start = 36  # '('
    arg_end = 37  # ')'
    randgen_marker = 38
    randgen_name = 39
    opposite_randgen_marker = 40
    percentgen_marker = 41
    percentgen = 42
    variation_marker = 43
    variation_name = 44


class LexicalToken(object):
    """Represents a terminal token with a given `TerminalType`."""
    def __init__(self, terminal_type, token_text):
        self.type = terminal_type
        if len(token_text) == 0 and (self.type != TerminalType.ignore):
            raise ValueError("Tried to create a lexed item of 0 characters.")
        self.text = token_text

    def __str__(self):
        return self.__repr__()
    def __repr__(self):
        return \
            "LexicalToken(type: " + self.type.name + \
            ", tokens: \"" + self.text + "\")"

    def remove_escapement(self):
        """Removes all the escapement from the text of this token."""
        processed_text = ""
        for i in range(len(self.text) - 1):
            if not (
                self.text[i] == ESCAPEMENT_SYM
                and self.text[i+1] in ESCAPABLE_CHARS
            ):
                processed_text += self.text[i]
        processed_text += self.text[-1]
        self.text = processed_text


def remove_comment_tokens(tokens):
    """
    Returns all the tokens except the ones that correspond to comments (and
    the preceding whitespaces if any).
    """
    if len(tokens) == 0:
        return tokens

    comment_index = None
    for (i, token) in enumerate(tokens):
        if token.type == TerminalType.comment:
            comment_index = i
            break

    if comment_index is None:
        return tokens
    if comment_index == 0:
        return []
    if tokens[comment_index - 1].type == TerminalType.whitespace:
        return tokens[:comment_index - 1]
    return tokens[:comment_index]


# TODO: unused function?
def extract_annotation_tokens(tokens):
    """
    Given the list of tokens `tokens`,
    returns a list of tokens that correspond to an annotation.
    Returns `None` if there was no annotation in `tokens`.
    """
    if len(tokens) < 2:
        return None

    start_index = None
    end_index = None
    for (i, token) in enumerate(tokens):
        if token.type == TerminalType.annotation_start:
            start_index = i
        if token.type == TerminalType.annotation_end:
            end_index = i
            break

    if start_index is None:
        return None
    if end_index is None:
        return tokens[start_index:]
    return tokens[start_index:end_index + 1]


def find_matching_choice_end(tokens, start_index):
    """
    Returns the index of the choice end that matches the choice start at
    index `start_index`.
    Returns `None` if there is no matching choice end.
    @raises: - `ValueError` if there is no choice starting at `start_index`.
             - `ValueError` if arguments are inconsistent.
    """
    if tokens[start_index].type != TerminalType.choice_start:
        raise ValueError(
            "Tried to get the matching choice end of something else than a " + \
            "choice start."
        )

    nb_starts_to_match = 0
    i = start_index + 1
    while i < len(tokens):
        token = tokens[i]
        if token.type == TerminalType.choice_start:
            nb_starts_to_match += 1
        elif token.type == TerminalType.choice_end:
            if nb_starts_to_match > 0:
                nb_starts_to_match -= 1
            else:
                return i
        i += 1
    return None

def find_index_last_choice_content(tokens, start_index):
    """
    Returns the index of the last token that makes up the internal rules
    of the choice starting at index `start_index`. In other words,
    the returned index points to the last token before the end of the choice
    or its random generation markers.
    Returns `None` if the choice is incorrectly closed.
    @raises: - `ValueError` if there isno choice starting at `start_index`.
    """
    end_choice_index = find_matching_choice_end(tokens, start_index)
    i = end_choice_index - 1
    if i > 0 and tokens[i].type == TerminalType.percentgen:
        i -= 1
    if i > 0 and tokens[i].type == TerminalType.percentgen_marker:
        i -= 1
    if i > 0 and tokens[i].type == TerminalType.randgen_name:
        i -= 1
    if i > 0 and tokens[i].type == TerminalType.opposite_randgen_marker:
        i -= 1
    if i > 0 and tokens[i].type == TerminalType.randgen_marker:
        i -= 1
    return i



"""
Module `chatette.cli.interactive_commands.command_strategy`.
Contains the base class for all the commands that can be used with
the interactive mode, such that such a command will subclass this base class
in order to implement the strategy design pattern.
"""

import re

from chatette.utils import rchop, UnitType
from chatette.parsing.utils import \
    ALIAS_SYM, SLOT_SYM, INTENT_SYM, VARIATION_SYM, ESCAPEMENT_SYM
from chatette.cli.terminal_writer import TerminalWriter, RedirectionType

from chatette.units.ast import AST


REDIRECTION_SYM = ">"
REDIRECTION_APPEND_SYM = ">>"

REGEX_SYM = '/'


class CommandStrategy(object):
    usage_str = "Undefined"  # Should be overriden by subclasses
    def __init__(self, command_str, quiet=False):
        self.command_tokens = CommandStrategy.tokenize(command_str)

        redirection_tuple = \
                CommandStrategy.find_redirection_file_path(self.command_tokens)
        if redirection_tuple is not None:
            self.remove_redirection_tokens()
            (redirection_type, redirection_filepath) = redirection_tuple
            self.print_wrapper = \
                TerminalWriter(redirection_type, redirection_filepath)
        elif quiet:
            self.print_wrapper = \
                TerminalWriter(redirection_type=RedirectionType.quiet)
        else:
            self.print_wrapper = TerminalWriter(None)

        self._is_regex_global = None

    @staticmethod
    def tokenize(command_str):
        """Tokenizes a string that is a command."""
        splitted = command_str.split()
        if len(splitted) == 0:
            return []
        if len(splitted) == 1:
            return [splitted[0]]
        tokens = []
        current_token = ""
        inside_token = False
        inside_regex = False
        for word in splitted:
            if inside_token:
                current_token += ' ' + word
                if word.endswith('"') and (len(word) < 2 or word[-2] != '\\'):
                    inside_token = False
                    tokens.append(current_token.rstrip())
                    current_token = ""
            elif inside_regex:
                current_token += ' ' + word
                if CommandStrategy._is_end_regex(word):
                    inside_regex = False
                    tokens.append(current_token)
                    current_token = ""
            elif (
                word.startswith('"') and word.endswith('"')
                and (len(word) < 2 or word[-2] != '\\')
            ):
                tokens.append(word)
            elif word.startswith('"'):
                inside_token = True
                current_token += word
            elif word.startswith("/") and CommandStrategy._is_end_regex(word):
                tokens.append(word)
            elif word.startswith("/"):
                inside_regex = True
                current_token += word
            else:  # not inside a token and not starting with "
                tokens.append(word)
        if current_token != "":
            tokens.append(current_token.rstrip())
        return tokens
    @staticmethod
    def _is_end_regex(word):
        """Returns `True` if `word` is the end of a regex."""
        return \
            word.endswith("/") or word.endswith("/g") \
            or word.endswith("/i") or word.endswith("/ig") \
            or word.endswith("/gi")

    @staticmethod
    def find_redirection_file_path(tokens):
        """
        Finds the path of the file
        which the output of a command should be redirected to and
        returns it along with the type of redirection that it is (in a 2-tuple).
        The type of redirection is an enumeration item of type
        `RedirectionType`.
        Returns `None` if no redirection was found.
        Returns `(RedirectionType.quiet, None)` if a redirection should be done
        to nowhere.
        (This can be achieved by using the redirections symbols and providing
        no filepath).
        """
        if len(tokens) < 2:
            return None
        if tokens[-2] == REDIRECTION_APPEND_SYM:
            return (RedirectionType.append, tokens[-1])
        if tokens[-2] == REDIRECTION_SYM:
            return (RedirectionType.truncate, tokens[-1])
        if (
            tokens[-1] == REDIRECTION_APPEND_SYM
            or tokens[-1] == REDIRECTION_SYM
        ):
            return (RedirectionType.quiet, None)
        return None

    @staticmethod
    def get_unit_type_from_str(unit_type_str):
        """
        Transforms the string `unit_type_str`
        into the corresponding `UnitType` value.
        Returns `None` if there exist no corresponding `UnitType` value.
        """
        unit_type_str = unit_type_str.lower()
        if unit_type_str in ("alias", ALIAS_SYM):
            return UnitType.alias
        if unit_type_str in ("slot", SLOT_SYM):
            return UnitType.slot
        if unit_type_str in ("intent", INTENT_SYM):
            return UnitType.intent
        return None


    @staticmethod
    def remove_quotes(text):
        """
        Removes the double quotes at the beginning and end of `text` and
        returns it. Double quotes that shouldn't be considered are escaped.
        """
        return text[1:-1].replace(ESCAPEMENT_SYM + '"', '"')

    @staticmethod
    def split_exact_unit_name(text):
        """
        Removes the double quotes at the beginning and end of `text` and
        splits `text` into a unit name and a variation identifier.
        Those two parts are separated in `text` by a hashtag `#` (unescaped).
        Returns a list with both those str inside it or the unit name and
        `None` for the variation identifier if it wasn't found.
        @raises: - `SyntaxError` if `text` is not valid (several hashtags),
        """
        no_quote_text = text[1:-1].replace(ESCAPEMENT_SYM + '"', '"')
        splitted = no_quote_text.split(VARIATION_SYM)
        processed_splitted = []
        current_word = ""
        for word in splitted:
            if word.endswith(ESCAPEMENT_SYM):
                if current_word == "":
                    current_word = word[:-1]
                else:
                    current_word += VARIATION_SYM + word[:-1]
            elif current_word == "":
                processed_splitted.append(word)
            else:
                processed_splitted.append(current_word + VARIATION_SYM + word)
                current_word = ""
        if len(processed_splitted) > 2:
            raise SyntaxError("Too many hashtags in unit identifier.")
        if len(processed_splitted) == 1:
            return [processed_splitted[0], None]
        return processed_splitted


    def get_regex_name(self, name):
        """
        If a regex is used within `name`,
        returns a compiled regex that is able to find all units with a name
        that matches some pattern.
        Returns `None` otherwise.
        """
        if (
            name.startswith(REGEX_SYM)
            and (
                name.endswith(REGEX_SYM)
                or name[-2] == REGEX_SYM
                or name[-3] == REGEX_SYM)
        ):
            splitted_str = [e for e in name.split(REGEX_SYM) if e != ""]
            if len(splitted_str) == 1:
                return re.compile(name[1:-1].replace('\\'+REGEX_SYM, REGEX_SYM))

            flags = splitted_str[-1]
            text_without_flags = rchop(name, flags)
            self._is_regex_global = 'g' in flags
            if 'i' in flags:
                return re.compile(
                    text_without_flags[1:-1].replace(
                        '\\'+REGEX_SYM, REGEX_SYM
                    ),
                    re.IGNORECASE
                )
            return re.compile(
                text_without_flags[1:-1].replace('\\'+REGEX_SYM, REGEX_SYM)
            )
        return None

    def next_matching_unit_name(self, unit_type, regex):
        """
        Yields the next unit name of type `unit_type`
        whose name matches `regex`.
        """
        relevant_dict = AST.get_or_create()._get_relevant_dict(unit_type)
        if self._is_regex_global:
            for unit_name in relevant_dict:
                if regex.search(unit_name):
                    yield unit_name
        else:
            for unit_name in relevant_dict:
                if regex.match(unit_name):
                    yield unit_name
    def get_all_matching_unit_names(self, unit_type, regex):
        """
        Returns a list of unit names of type `unit_type`
        whose name matches `regex`.
        NOTE: this is used with 'delete' command since the generator that
              returns the same data cannot be used in that case (dict changes
              size during iteration).
        """
        relevant_dict = AST.get_or_create()[unit_type]
        if self._is_regex_global:
            return [name for name in relevant_dict if regex.search(name)]
        return [name for name in relevant_dict if regex.match(name)]

    def remove_redirection_tokens(self):
        """
        Removes the tokens that represent a redirection
        from `self.command_tokens`.
        @pre: There are redirection tokens in the tokens.
        """
        if (
            self.command_tokens[-2] == REDIRECTION_APPEND_SYM
            or self.command_tokens[-2] == REDIRECTION_SYM
        ):
            self.command_tokens = self.command_tokens[:-2]
        else:
            self.command_tokens = self.command_tokens[:-1]


    def flush_output(self):
        """
        Asks the wrapper of print to flush its outputs
        to the redirected file (if such a file exists).
        """
        self.print_wrapper.flush()


    def should_exit(self):
        """
        Returns `True` if the program should exit the interactive mode.
        This method should be overriden by subclasses.
        """
        return False

    def execute(self):
        """
        Executes the whole command represented by this object.
        This method can be overriden by subclasses if a different algorithm is
        required.
        """
        # TODO support variations
        if len(self.command_tokens) < 3:
            self.print_wrapper.error_log(
                "Missing some arguments\nUsage: " + self.usage_str
            )
            return

        unit_type = CommandStrategy.get_unit_type_from_str(self.command_tokens[1])
        if unit_type is None:
            self.print_wrapper.error_log(
                "Unknown unit type: '" + str(self.command_tokens[1]) + "'."
            )
            return

        unit_regex = self.get_regex_name(self.command_tokens[2])
        if unit_regex is None:
            try:
                [unit_name, variation_name] = \
                    CommandStrategy.split_exact_unit_name(
                        self.command_tokens[2]
                    )
            except SyntaxError:
                self.print_wrapper.error_log(
                    "Unit identifier couldn't be interpreted. " + \
                    "Did you mean to escape some hashtags '#'?")
                return
            self.execute_on_unit(unit_type, unit_name, variation_name)
        else:
            count = 0
            for unit_name in self.next_matching_unit_name(
                unit_type, unit_regex
            ):
                self.execute_on_unit(unit_type, unit_name)
                count += 1
            if count == 0:
                self.print_wrapper.write("No " + unit_type.name + " matched.")
        self.finish_execution()

    def execute_on_unit(self, unit_type, unit_name, variation_name=None):
        """
        Executes the command on a specific unit.
        This method HAS to be overriden by subclasses if they don't override
        `execute`.
        """
        raise NotImplementedError()

    def finish_execution(self):
        """
        This function is executed at the end of the command and
        can be overriden to implement things such as cleaning up or changing
        the size of dictionaries and list that were being iterated over.
        NOTE: it is called no matter what happened before (the command worked
              or failed).
        """
        pass



from .pyversion import _is_supported_python_version, _get_python_version_as_str


class PreconditionsUnmet(Exception):
    pass


def ensure_preconditions():
    _ensure_python_version_is_supported()

def _ensure_python_version_is_supported():
    if not _is_supported_python_version():
        print(
            "[ERROR] Python v" + _get_python_version_as_str() + \
            " is not supported by Chatette. " + \
            "Please use a version of Python older than v2.7.\n"
        )
        raise PreconditionsUnmet()
